#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the logging utilities for restate handlers.
"""

import logging

from .server_context import restate_context_is_replaying


# pylint: disable=C0103
def getLogger(name=None):
    """
    Wrapper for logging.getLogger returning a logger configured with RestateLoggingFilter

    :param name: the logger name, if any
    :return: the logger as returned by logging.getLogger, configured with the RestateLoggingFilter
    """
    logger = logging.getLogger(name)
    logger.addFilter(RestateLoggingFilter())
    return logger


# pylint: disable=R0903
class RestateLoggingFilter(logging.Filter):
    """
    Restate logging filter. This filter will filter out logs on replay
    """

    def filter(self, record):
        # First, apply the filter base logic
        if not super().filter(record):
            return False

        # Read the context variable, check if we're replaying
        if restate_context_is_replaying.get():
            return False

        # We're not replaying, all good pass the event
        return True



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
Restate SDK for Python
"""

from contextlib import asynccontextmanager
import typing

from restate.server_types import RestateAppT
from restate.types import HarnessEnvironment

from .service import Service
from .object import VirtualObject
from .workflow import Workflow

# types
from .context import Context, ObjectContext, ObjectSharedContext
from .context import WorkflowContext, WorkflowSharedContext
from .retry_policy import InvocationRetryPolicy
from .client_types import RestateClient, HttpError, RestateClientSendHandle

# pylint: disable=line-too-long
from .context import (
    DurablePromise,
    RestateDurableFuture,
    RestateDurableCallFuture,
    RestateDurableSleepFuture,
    SendHandle,
    RunOptions,
)
from .exceptions import TerminalError, SdkInternalBaseException, is_internal_exception
from .asyncio import as_completed, gather, wait_completed, select

from .endpoint import app

from .logging import getLogger, RestateLoggingFilter

try:
    from .harness import create_test_harness, test_harness  # type: ignore
except ImportError:
    # we don't have the appropriate dependencies installed

    # pylint: disable=unused-argument, redefined-outer-name
    @asynccontextmanager
    def create_test_harness(
        app: RestateAppT,
        follow_logs: bool = False,
        restate_image: str = "docker.io/restatedev/restate:latest",
        always_replay: bool = False,
        disable_retries: bool = False,
    ) -> typing.AsyncGenerator[HarnessEnvironment, None]:
        """a dummy harness constructor to raise ImportError. Install restate-sdk[harness] to use this feature"""
        raise ImportError("Install restate-sdk[harness] to use this feature")

    @typing.no_type_check
    def test_harness(
        app: RestateAppT,
        follow_logs: bool = False,
        restate_image: str = "docker.io/restatedev/restate:latest",
        always_replay: bool = False,
        disable_retries: bool = False,
    ):
        """a dummy harness constructor to raise ImportError. Install restate-sdk[harness] to use this feature"""
        raise ImportError("Install restate-sdk[harness] to use this feature")  # type: ignore


try:
    from .client import create_client
except ImportError:
    # we don't have the appropriate dependencies installed

    @asynccontextmanager
    async def create_client(
        ingress: str, headers: typing.Optional[dict] = None
    ) -> typing.AsyncGenerator[RestateClient, None]:
        """a dummy client constructor to raise ImportError. Install restate-sdk[client] to use this feature"""
        raise ImportError("Install restate-sdk[client] to use this feature")
        yield  # type: ignore


__all__ = [
    "Service",
    "VirtualObject",
    "Workflow",
    "Context",
    "ObjectContext",
    "ObjectSharedContext",
    "WorkflowContext",
    "WorkflowSharedContext",
    "DurablePromise",
    "RestateDurableFuture",
    "RestateDurableCallFuture",
    "RestateDurableSleepFuture",
    "SendHandle",
    "RunOptions",
    "TerminalError",
    "app",
    "create_test_harness",
    "test_harness",
    "HarnessEnvironment",
    "gather",
    "as_completed",
    "wait_completed",
    "select",
    "logging",
    "RestateLoggingFilter",
    "InvocationRetryPolicy",
    "SdkInternalBaseException",
    "is_internal_exception",
    "getLogger",
    "RestateClient",
    "RestateClientSendHandle",
    "HttpError",
    "create_client",
]



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""

from datetime import datetime

# pylint: disable=C0116
# pylint: disable=W0613

from restate import Service, Context

random_greeter = Service("random_greeter")


@random_greeter.handler()
async def greet(ctx: Context, name: str) -> str:
    # ctx.random() returns a Python Random instance seeded deterministically.
    # By using ctx.random() you don't write entries in the journal,
    # but you still get the same generated values on retries.

    # To generate random numbers
    random_number = ctx.random().randint(0, 100)

    # To generate random bytes
    random_bytes = ctx.random().randbytes(10)

    # Use ctx.uuid() to generate a UUID v4 seeded deterministically
    # As with ctx.random(), this won't write entries in the journal
    random_uuid = ctx.uuid()

    # To get a timestamp, use ctx.time()
    # This will record the timestamp in the Restate journal
    now = await ctx.time()

    # You can convert it to date/datetime using Python's standard library functions, e.g.
    now_datetime = datetime.fromtimestamp(now)

    # Or to perform a difference:
    # start = await ctx.time()
    # # Some code
    # end = await ctx.time()
    # delta = datetime.timedelta(seconds=(end-start))

    return (
        f"Hello {name} with "
        f"random number {random_number}, "
        f"random bytes {random_bytes!r} "
        f"random uuid {random_uuid},"
        f"now datetime {now_datetime}!"
    )



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from datetime import timedelta
from typing import Dict, List
from restate import Service, Context
from restate.serde import BytesSerde

test_utils = Service("TestUtilsService")


@test_utils.handler()
async def echo(context: Context, input: str) -> str:
    return input


@test_utils.handler(name="uppercaseEcho")
async def uppercase_echo(context: Context, input: str) -> str:
    return input.upper()


@test_utils.handler(name="echoHeaders")
async def echo_headers(context: Context) -> Dict[str, str]:
    return context.request().headers


@test_utils.handler(
    name="rawEcho",
    accept="*/*",
    content_type="application/octet-stream",
    input_serde=BytesSerde(),
    output_serde=BytesSerde(),
)
async def raw_echo(context: Context, input: bytes) -> bytes:
    return input


@test_utils.handler(name="sleepConcurrently")
async def sleep_concurrently(context: Context, millis_duration: List[int]) -> None:
    timers = [context.sleep(timedelta(milliseconds=duration)) for duration in millis_duration]

    for timer in timers:
        await timer


@test_utils.handler(name="countExecutedSideEffects")
async def count_executed_side_effects(context: Context, increments: int) -> int:
    invoked_side_effects = 0

    def effect():
        nonlocal invoked_side_effects
        invoked_side_effects += 1

    for _ in range(increments):
        await context.run("count", effect)

    return invoked_side_effects


@test_utils.handler(name="cancelInvocation")
async def cancel_invocation(context: Context, invocation_id: str) -> None:
    context.cancel_invocation(invocation_id)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

# pylint: disable=R0917
"""
This module contains the definition of the Handler class,
which is used to define the handlers for the services.
"""

from dataclasses import dataclass
from datetime import timedelta
from inspect import Signature
from typing import (
    Any,
    AsyncContextManager,
    Callable,
    Awaitable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    TypeVar,
)

from restate.retry_policy import InvocationRetryPolicy

from restate.context import HandlerType
from restate.exceptions import TerminalError
from restate.serde import DefaultSerde, PydanticJsonSerde, MsgspecJsonSerde, Serde, is_pydantic, Msgspec
from restate.types import extract_core_type

I = TypeVar("I")
O = TypeVar("O")
T = TypeVar("T")

# we will use this symbol to store the handler in the function
RESTATE_UNIQUE_HANDLER_SYMBOL = str(object())


@dataclass
class ServiceTag:
    """
    This class is used to identify the service.
    """

    kind: Literal["object", "service", "workflow"]
    name: str
    description: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None


@dataclass
class TypeHint(Generic[T]):
    """
    Represents a type hint.
    """

    annotation: Optional[T] = None
    """The type annotation."""
    is_void: bool = False
    """Whether the type is void (i.e., None)."""
    generate_json_schema: Callable[[], Dict[str, Any]] | None = None
    """A callable that generates the JSON schema for the type."""


@dataclass
class HandlerIO(Generic[I, O]):
    """
    Represents the input/output configuration for a handler.

    Attributes:
        accept (str): The accept header value for the handler.
        content_type (str): The content type header value for the handler.
    """

    accept: str
    content_type: str
    input_serde: Serde[I]
    output_serde: Serde[O]
    input_type: Optional[TypeHint[I]] = None
    output_type: Optional[TypeHint[O]] = None


def _json_schema_wrap_as_optional(schema: Dict[str, Any]) -> Dict[str, Any]:
    """
    modify the given JSON schema with its type wrapped as optional (nullable).
    """
    t = schema.get("type")

    if t is None:
        # If type is unspecified, leave it open by only adding "null"
        schema["type"] = ["null"]
        return schema

    if isinstance(t, list):
        if "null" not in t:
            t.append("null")
    else:
        if t != "null":
            schema["type"] = [t, "null"]

    return schema


def _make_json_schema_generator(
    original: Callable[[], Dict[str, Any]], type: Literal["optional", "simple"]
) -> Callable[[], Dict[str, Any]]:
    """
    Create a JSON schema generator that handles optional types.

    If the type is optional, the generated schema will include "null" in the type.
    """
    if type == "simple":
        return original

    def generator() -> Dict[str, Any]:
        schema = original()
        if type == "optional":
            return _json_schema_wrap_as_optional(schema)

        assert False, "unreachable"

    return generator


def update_handler_io_with_input_type_hints(handler_io: HandlerIO[I, O], signature: Signature):
    """
    Augment handler_io with additional information about the input type.

    This function has a special check for msgspec Structs and Pydantic models when these are provided.
    This method will inspect the signature of an handler and will look for
    the input type of a function, and will:
    * capture any msgspec Structs or Pydantic models (to be used later at discovery)
    * replace the default json serializer (is unchanged by a user) with the appropriate serde
    """
    params = list(signature.parameters.values())
    if len(params) == 1:
        # if there is only one parameter, it is the context.
        handler_io.input_type = TypeHint(is_void=True)
        return

    annotation = params[-1].annotation
    core_kind, core_type = extract_core_type(annotation)
    handler_io.input_type = TypeHint(annotation=core_type)
    if Msgspec.is_struct(core_type):
        handler_io.input_type.generate_json_schema = _make_json_schema_generator(
            lambda: Msgspec.json_schema(core_type), core_kind
        )
        if isinstance(handler_io.input_serde, DefaultSerde):
            handler_io.input_serde = MsgspecJsonSerde(core_type)
        return

    if is_pydantic(core_type):
        handler_io.input_type.generate_json_schema = _make_json_schema_generator(
            lambda: core_type.model_json_schema(mode="serialization"), core_kind
        )
        if isinstance(handler_io.input_serde, DefaultSerde):
            handler_io.input_serde = PydanticJsonSerde(core_type)


def update_handler_io_with_return_type_hints(handler_io: HandlerIO[I, O], signature: Signature):
    """
    Augment handler_io with additional information about the output type.

    This function has a special check for msgspec Structs and Pydantic models when these are provided.
    This method will inspect the signature of an handler and will look for
    the return type of a function, and will:
    * capture any msgspec Structs or Pydantic models (to be used later at discovery)
    * replace the default json serializer (is unchanged by a user) with the appropriate serde
    """
    return_annotation = signature.return_annotation
    if return_annotation is None or return_annotation is Signature.empty:
        # if there is no return annotation, we assume it is void
        handler_io.output_type = TypeHint(is_void=True)
        return

    core_kind, return_core_type = extract_core_type(return_annotation)
    handler_io.output_type = TypeHint(annotation=return_core_type)
    if Msgspec.is_struct(return_core_type):
        handler_io.output_type.generate_json_schema = _make_json_schema_generator(
            lambda: Msgspec.json_schema(return_core_type), core_kind
        )
        if isinstance(handler_io.output_serde, DefaultSerde):
            handler_io.output_serde = MsgspecJsonSerde(return_core_type)
        return

    if is_pydantic(return_core_type):
        handler_io.output_type.generate_json_schema = _make_json_schema_generator(
            lambda: return_core_type.model_json_schema(mode="serialization"), core_kind
        )
        if isinstance(handler_io.output_serde, DefaultSerde):
            handler_io.output_serde = PydanticJsonSerde(return_core_type)


def update_handler_io_with_type_hints(handler_io: HandlerIO[I, O], signature: Signature):
    """
    Augment handler_io with additional information about the input and output types.

    This function has a special check for msgspec Structs and Pydantic models when these are provided.
    This method will inspect the signature of an handler and will look for
    the input and the return types of a function, and will:
    * capture any msgspec Structs or Pydantic models (to be used later at discovery)
    * replace the default json serializer (is unchanged by a user) with the appropriate serde
    """
    update_handler_io_with_input_type_hints(handler_io, signature)
    update_handler_io_with_return_type_hints(handler_io, signature)


# pylint: disable=R0902
@dataclass
class Handler(Generic[I, O]):
    """
    Represents a handler for a service.

    Attributes:
        service_tag: The service tag for the handler.
        handler_io: The input/output configuration for the handler.
        kind: The kind of handler (exclusive, shared, workflow).
        name: The name of the handler.
        fn: The handler function.
        arity: The number of parameters in the handler function.
        description: Documentation for this handler definition.
        metadata: Custom metadata for this handler definition.
        inactivity_timeout: Inactivity timeout duration.
        abort_timeout: Abort timeout duration.
        journal_retention: Journal retention duration.
        idempotency_retention: Idempotency retention duration.
        workflow_retention: Workflow completion retention duration.
        enable_lazy_state: If true, lazy state is enabled.
        ingress_private: If true, the handler cannot be invoked from the HTTP nor Kafka ingress.
        invocation_retry_policy: Optional retry policy configuration applied to this handler.
    """

    service_tag: ServiceTag
    handler_io: HandlerIO[I, O]
    kind: Optional[Literal["exclusive", "shared", "workflow"]]
    name: str
    fn: Callable[[Any, I], Awaitable[O]] | Callable[[Any], Awaitable[O]]
    arity: int
    description: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    inactivity_timeout: Optional[timedelta] = None
    abort_timeout: Optional[timedelta] = None
    journal_retention: Optional[timedelta] = None
    idempotency_retention: Optional[timedelta] = None
    workflow_retention: Optional[timedelta] = None
    enable_lazy_state: Optional[bool] = None
    ingress_private: Optional[bool] = None
    invocation_retry_policy: Optional[InvocationRetryPolicy] = None
    context_managers: Optional[List[Callable[[], AsyncContextManager[None]]]] = None


# disable too many arguments warning
# pylint: disable=R0913
# pylint: disable=R0914
def make_handler(
    service_tag: ServiceTag,
    handler_io: HandlerIO[I, O],
    name: str | None,
    kind: Optional[Literal["exclusive", "shared", "workflow"]],
    wrapped: Any,
    signature: Signature,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, str]] = None,
    inactivity_timeout: Optional[timedelta] = None,
    abort_timeout: Optional[timedelta] = None,
    journal_retention: Optional[timedelta] = None,
    idempotency_retention: Optional[timedelta] = None,
    workflow_retention: Optional[timedelta] = None,
    enable_lazy_state: Optional[bool] = None,
    ingress_private: Optional[bool] = None,
    invocation_retry_policy: Optional[InvocationRetryPolicy] = None,
    context_managers: Optional[List[Callable[[], AsyncContextManager[None]]]] = None,
) -> Handler[I, O]:
    """
    Factory function to create a handler.

    Args:
        service_tag: The service tag for the handler.
        handler_io: The input/output configuration for the handler.
        name: The name of the handler.
        kind: The kind of handler (exclusive, shared, workflow).
        wrapped: The wrapped function.
        signature: The signature of the function.
        description: Documentation for this handler definition.
        metadata: Custom metadata for this handler definition.
        inactivity_timeout: Inactivity timeout duration.
        abort_timeout: Abort timeout duration.
        journal_retention: Journal retention duration.
        idempotency_retention: Idempotency retention duration.
        workflow_retention: Workflow completion retention duration.
        enable_lazy_state: If true, lazy state is enabled.
        ingress_private: If true, the handler cannot be invoked from the HTTP nor Kafka ingress.
        invocation_retry_policy: Retry policy used by Restate when invoking this handler.
            NOTE: only supported with restate-server >= 1.5.
    """
    # try to deduce the handler name
    handler_name = name
    if not handler_name:
        handler_name = wrapped.__name__
    if not handler_name:
        raise ValueError("Handler name must be provided")

    if len(signature.parameters) == 0:
        raise ValueError("Handler must have at least one parameter")

    arity = len(signature.parameters)
    update_handler_io_with_type_hints(handler_io, signature)  # mutates handler_io

    handler = Handler[I, O](
        service_tag=service_tag,
        handler_io=handler_io,
        kind=kind,
        name=handler_name,
        fn=wrapped,
        arity=arity,
        description=description,
        metadata=metadata,
        inactivity_timeout=inactivity_timeout,
        abort_timeout=abort_timeout,
        journal_retention=journal_retention,
        idempotency_retention=idempotency_retention,
        workflow_retention=workflow_retention,
        enable_lazy_state=enable_lazy_state,
        ingress_private=ingress_private,
        invocation_retry_policy=invocation_retry_policy,
        context_managers=context_managers,
    )

    vars(wrapped)[RESTATE_UNIQUE_HANDLER_SYMBOL] = handler
    return handler


def handler_from_callable(wrapper: HandlerType[I, O]) -> Handler[I, O]:
    """
    Get the handler from the callable.
    """
    try:
        return vars(wrapper)[RESTATE_UNIQUE_HANDLER_SYMBOL]
    except KeyError:
        raise ValueError("Handler not found")  # pylint: disable=raise-missing-from


async def invoke_handler(handler: Handler[I, O], ctx: Any, in_buffer: bytes) -> bytes:
    """
    Invoke the handler with the given context and input.
    """
    if handler.arity == 2:
        try:
            in_arg = handler.handler_io.input_serde.deserialize(in_buffer)
        except Exception as e:
            raise TerminalError(message=f"Unable to parse an input argument. {e}") from e
        out_arg = await handler.fn(ctx, in_arg)  # type: ignore [call-arg, arg-type]
    else:
        out_arg = await handler.fn(ctx)  # type: ignore [call-arg]
    out_buffer = handler.handler_io.output_serde.serialize(out_arg)
    return bytes(out_buffer)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
import restate

from virtual_object import increment, count

#
# uv run examples/client.py
#


async def main():
    async with restate.create_client("http://localhost:8080") as client:
        await client.object_call(increment, key="a", arg=5)

        await client.object_send(increment, key="a", arg=5)

        current_count = await client.object_call(count, key="a", arg=None)

        print(f"Current count for 'a': {current_count}")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""

from datetime import timedelta

# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=W0622

from restate import VirtualObject, ObjectContext
from restate.exceptions import TerminalError
from restate import RunOptions

failing = VirtualObject("Failing")


@failing.handler(name="terminallyFailingCall")
async def terminally_failing_call(ctx: ObjectContext, msg: str):
    raise TerminalError(message=msg)


@failing.handler(name="callTerminallyFailingCall")
async def call_terminally_failing_call(ctx: ObjectContext, msg: str) -> str:
    await ctx.object_call(terminally_failing_call, key="random-583e1bf2", arg=msg)

    raise Exception("Should not reach here")


failures = 0


@failing.handler(name="failingCallWithEventualSuccess")
async def failing_call_with_eventual_success(ctx: ObjectContext) -> int:
    global failures
    failures += 1
    if failures >= 4:
        failures = 0
        return 4
    raise ValueError(f"Failed at attempt: {failures}")


@failing.handler(name="terminallyFailingSideEffect")
async def terminally_failing_side_effect(ctx: ObjectContext, error_message: str):
    def side_effect():
        raise TerminalError(message=error_message)

    await ctx.run_typed("sideEffect", side_effect)
    raise ValueError("Should not reach here")


eventual_success_side_effects = 0


@failing.handler(name="sideEffectSucceedsAfterGivenAttempts")
async def side_effect_succeeds_after_given_attempts(ctx: ObjectContext, minimum_attempts: int) -> int:
    def side_effect():
        global eventual_success_side_effects
        eventual_success_side_effects += 1
        if eventual_success_side_effects >= minimum_attempts:
            return eventual_success_side_effects
        raise ValueError(f"Failed at attempt: {eventual_success_side_effects}")

    options: RunOptions[int] = RunOptions(
        max_attempts=minimum_attempts + 1, initial_retry_interval=timedelta(milliseconds=1), retry_interval_factor=1.0
    )
    return await ctx.run_typed("sideEffect", side_effect, options)


eventual_failure_side_effects = 0


@failing.handler(name="sideEffectFailsAfterGivenAttempts")
async def side_effect_fails_after_given_attempts(ctx: ObjectContext, retry_policy_max_retry_count: int) -> int:
    def side_effect():
        global eventual_failure_side_effects
        eventual_failure_side_effects += 1
        raise ValueError(f"Failed at attempt: {eventual_failure_side_effects}")

    try:
        options: RunOptions[int] = RunOptions(
            max_attempts=retry_policy_max_retry_count,
            initial_retry_interval=timedelta(milliseconds=1),
            retry_interval_factor=1.0,
        )
        await ctx.run_typed("sideEffect", side_effect, options)
        raise ValueError("Side effect did not fail.")
    except TerminalError:
        global eventual_failure_side_effects
        return eventual_failure_side_effects



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains functions for serializing and deserializing data."""

import abc
import json
import typing

from dataclasses import asdict, is_dataclass

T = typing.TypeVar("T")
I = typing.TypeVar("I")
O = typing.TypeVar("O")


def try_import_pydantic_base_model():
    """
    Try to import PydanticBaseModel from Pydantic.
    """
    try:
        from pydantic import BaseModel  # type: ignore # pylint: disable=import-outside-toplevel

        return BaseModel
    except ImportError:

        class Dummy:  # pylint: disable=too-few-public-methods
            """a dummy class to use when Pydantic is not available"""

        return Dummy


def try_import_from_dacite():
    """
    Try to import from_dict from dacite.
    """
    try:
        from dacite import from_dict  # type: ignore # pylint: disable=import-outside-toplevel

        def _to_dict(obj: typing.Any) -> dict[typing.Any, typing.Any]:
            return asdict(obj)

        def _from_dict(data_class: typing.Any, data: typing.Any) -> typing.Any:
            return from_dict(data_class, data)

        return _to_dict, _from_dict

    except ImportError:

        def _to_dict(obj: typing.Any) -> dict[typing.Any, typing.Any]:
            """a dummy function when dacite is not available"""
            raise RuntimeError(
                "Trying to deserialize into a @dataclass."
                "Please add the optional dependencies needed."
                "use pip install restate-sdk[serde] "
                "or"
                " pip install restate-sdk[all] to install all dependencies."
            )

        def _from_dict(data_class: typing.Any, data: typing.Any) -> typing.Any:  # pylint: disable=too-few-public-methods,unused-argument
            """a dummy function when dacite is not available"""

            raise RuntimeError(
                "Trying to deserialize into a @dataclass."
                "Please add the optional dependencies needed."
                "use pip install restate-sdk[serde] "
                "or"
                " pip install restate-sdk[all] to install all dependencies."
            )

        return _to_dict, _from_dict


class MsgspecJsonAPI:
    def is_struct(self, annotation: typing.Any) -> bool:
        return False

    def decode(self, buf: bytes, type: typing.Type[T]) -> T:
        raise NotImplementedError("Please use msgspec as a conditional dependency to use msgspec features.")

    def encode(self, obj: typing.Any) -> bytes:
        raise NotImplementedError("Please use msgspec as a conditional dependency to use msgspec features.")

    def json_schema(self, type: typing.Type[T]) -> dict[str, typing.Any]:
        raise NotImplementedError("Please use msgspec as a conditional dependency to use msgspec features.")


def try_import_msgspec_api():
    """
    Try to import msgspec API.
    """
    try:
        from msgspec import Struct  # type: ignore # pylint: disable=import-outside-toplevel
        import msgspec

        class MsgspecImpl(MsgspecJsonAPI):
            def is_struct(self, annotation: typing.Any) -> bool:
                try:
                    return issubclass(annotation, Struct)
                except TypeError:
                    # annotation is not a class or a type
                    return False

            def decode(self, buf: bytes, type: typing.Type[T]) -> T:
                return msgspec.json.decode(buf, type=type)

            def encode(self, obj: typing.Any) -> bytes:
                return msgspec.json.encode(obj)

            def json_schema(self, type: typing.Type[T]) -> dict[str, typing.Any]:
                return msgspec.json.schema(type)

        return MsgspecImpl()

    except ImportError:
        return MsgspecJsonAPI()


PydanticBaseModel = try_import_pydantic_base_model()
Msgspec = try_import_msgspec_api()
# pylint: disable=C0103
DaciteToDict, DaciteFromDict = try_import_from_dacite()

# disable to few parameters
# pylint: disable=R0903


def is_pydantic(annotation) -> bool:
    """
    Check if an object is a Pydantic model.
    """
    try:
        return issubclass(annotation, PydanticBaseModel)
    except TypeError:
        # annotation is not a class or a type
        return False


class Serde(typing.Generic[T], abc.ABC):
    """serializer/deserializer interface."""

    @abc.abstractmethod
    def deserialize(self, buf: bytes) -> typing.Optional[T]:
        """
        Deserializes a bytearray to an object.
        """

    @abc.abstractmethod
    def serialize(self, obj: typing.Optional[T]) -> bytes:
        """
        Serializes an object to a bytearray.
        """


class BytesSerde(Serde[bytes]):
    """A pass-trough serializer/deserializer."""

    def deserialize(self, buf: bytes) -> typing.Optional[bytes]:
        """
        Deserializes a bytearray to a bytearray.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[bytes]: The deserialized bytearray.
        """
        return buf

    def serialize(self, obj: typing.Optional[bytes]) -> bytes:
        """
        Serializes a bytearray to a bytearray.

        Args:
            obj (bytes): The bytearray to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()
        return obj


class JsonSerde(Serde[I]):
    """A JSON serializer/deserializer."""

    def deserialize(self, buf: bytes) -> typing.Optional[I]:
        """
        Deserializes a bytearray to a JSON object.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[I]: The deserialized JSON object.
        """
        if not buf:
            return None
        return json.loads(buf)

    def serialize(self, obj: typing.Optional[I]) -> bytes:
        """
        Serializes a JSON object to a bytearray.

        Args:
            obj (I): The JSON object to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()

        return bytes(json.dumps(obj), "utf-8")


class DefaultSerde(Serde[I]):
    """
    The default serializer/deserializer used when no explicit type hints are provided.

    Behavior:
    - Serialization:
        - If the object is an instance of Pydantic's `BaseModel`,
            it uses `model_dump_json()` for serialization.
        - Otherwise, it falls back to `json.dumps()`.
    - Deserialization:
        - Uses `json.loads()` to convert byte arrays into Python objects.
        - Does **not** automatically reconstruct Pydantic models;
            deserialized objects remain as generic JSON structures (dicts, lists, etc.).

    Serde Selection:
    - When using the `@handler` decorator, if a function's type hints specify a Pydantic model,
      `PydanticJsonSerde` is automatically selected instead of `DefaultSerde`.
    - `DefaultSerde` is only used if no explicit type hints are provided.

    This serde ensures compatibility with both structured (Pydantic) and unstructured JSON data,
    while allowing automatic serde selection based on type hints.
    """

    def __init__(self, type_hint: typing.Optional[typing.Type[I]] = None):
        super().__init__()
        self.type_hint = type_hint

    def with_maybe_type(self, type_hint: typing.Type[I] | None = None) -> "DefaultSerde[I]":
        """
        Returns a new instance of DefaultSerde with the provided type hint.
        This is useful for creating a serde that is specific to a certain type.
        NOTE: This method does not modify the current instance.
        Args:
            type_hint (Type[I] | None): The type hint to use for serialization/deserialization.
        Returns:
            DefaultSerde[I]: A new instance of DefaultSerde with the provided type hint.
        """
        return DefaultSerde(type_hint)

    def deserialize(self, buf: bytes) -> typing.Optional[I]:
        """
        Deserializes a byte array into a Python object.

        Args:
            buf (bytes): The byte array to deserialize.

        Returns:
            Optional[I]: The resulting Python object, or None if the input is empty.
        """
        if not buf:
            return None
        hint = self.type_hint
        if not hint:
            return json.loads(buf)
        if Msgspec.is_struct(hint):
            return Msgspec.decode(buf, type=hint)
        if is_pydantic(hint):
            return hint.model_validate_json(buf)  # type: ignore
        if is_dataclass(hint):
            data = json.loads(buf)
            return DaciteFromDict(hint, data)
        # although we have a type hint, we fall back to json.loads because we were not able to
        # identify a specific deserialization method, perhaps the user specified a default type
        # for another reason than serialization/deserialization.
        return json.loads(buf)

    def serialize(self, obj: typing.Optional[I]) -> bytes:
        """
        Serializes a Python object into a byte array.
        If the object is a msgspec Struct or Pydantic BaseModel, uses their respective methods.

        Args:
            obj (Optional[I]): The Python object to serialize.

        Returns:
            bytes: The serialized byte array.
        """
        if obj is None:
            return bytes()
        hint = self.type_hint
        if not hint:
            return json.dumps(obj).encode("utf-8")
        if Msgspec.is_struct(hint):
            return Msgspec.encode(obj)
        if is_pydantic(hint):
            return obj.model_dump_json().encode("utf-8")  # type: ignore[attr-defined]
        if is_dataclass(obj):
            data = DaciteToDict(obj)  # type: ignore
            return json.dumps(data).encode("utf-8")
        # although we have a type hint, we fall back to json.dumps because we were not able to
        # identify a specific serialization method, perhaps the user specified a default type
        # for another reason than serialization/deserialization.
        return json.dumps(obj).encode("utf-8")


class PydanticJsonSerde(Serde[I]):
    """
    Serde for Pydantic models to/from JSON
    """

    def __init__(self, model):
        self.model = model

    def deserialize(self, buf: bytes) -> typing.Optional[I]:
        """
        Deserializes a bytearray to a Pydantic model.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[I]: The deserialized Pydantic model.
        """
        if not buf:
            return None
        return self.model.model_validate_json(buf)

    def serialize(self, obj: typing.Optional[I]) -> bytes:
        """
        Serializes a Pydantic model to a bytearray.

        Args:
            obj (I): The Pydantic model to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()
        json_str = obj.model_dump_json()  # type: ignore[attr-defined]
        return json_str.encode("utf-8")


class MsgspecJsonSerde(Serde[I]):
    """
    Serde for msgspec Structs to/from JSON
    """

    def __init__(self, model):
        self.model = model

    def deserialize(self, buf: bytes) -> typing.Optional[I]:
        """
        Deserializes a bytearray to a msgspec Struct.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[I]: The deserialized msgspec Struct.
        """
        if not buf:
            return None

        return Msgspec.decode(buf, type=self.model)

    def serialize(self, obj: typing.Optional[I]) -> bytes:
        """
        Serializes a msgspec Struct to a bytearray.

        Args:
            obj (I): The msgspec Struct to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()
        return Msgspec.encode(obj)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from datetime import timedelta
from restate import Service, Context
from typing import TypedDict, Optional, Iterable

proxy = Service("Proxy")


class ProxyRequest(TypedDict):
    serviceName: str
    virtualObjectKey: Optional[str]
    handlerName: str
    message: Iterable[int]
    delayMillis: Optional[int]
    idempotencyKey: Optional[str]


@proxy.handler()
async def call(ctx: Context, req: ProxyRequest) -> Iterable[int]:
    response = await ctx.generic_call(
        req["serviceName"],
        req["handlerName"],
        bytes(req["message"]),
        req.get("virtualObjectKey"),
        req.get("idempotencyKey"),
    )
    return list(response)


@proxy.handler(name="oneWayCall")
async def one_way_call(ctx: Context, req: ProxyRequest) -> str:
    send_delay = None
    delayMillis = req.get("delayMillis")
    if delayMillis is not None:
        send_delay = timedelta(milliseconds=delayMillis)
    handle = ctx.generic_send(
        req["serviceName"],
        req["handlerName"],
        bytes(req["message"]),
        req.get("virtualObjectKey"),
        send_delay=send_delay,
        idempotency_key=req.get("idempotencyKey"),
    )
    invocation_id = await handle.invocation_id()
    return invocation_id


class ManyCallRequest(TypedDict):
    proxyRequest: ProxyRequest
    oneWayCall: bool
    awaitAtTheEnd: bool


@proxy.handler(name="manyCalls")
async def many_calls(ctx: Context, requests: Iterable[ManyCallRequest]):
    to_await = []

    for req in requests:
        if req["oneWayCall"]:
            send_delay = None
            delayMillis = req["proxyRequest"].get("delayMillis")
            if delayMillis is not None:
                send_delay = timedelta(milliseconds=delayMillis)
            ctx.generic_send(
                req["proxyRequest"]["serviceName"],
                req["proxyRequest"]["handlerName"],
                bytes(req["proxyRequest"]["message"]),
                req["proxyRequest"].get("virtualObjectKey"),
                send_delay=send_delay,
                idempotency_key=req["proxyRequest"].get("idempotencyKey"),
            )
        else:
            awaitable = ctx.generic_call(
                req["proxyRequest"]["serviceName"],
                req["proxyRequest"]["handlerName"],
                bytes(req["proxyRequest"]["message"]),
                req["proxyRequest"].get("virtualObjectKey"),
                idempotency_key=req["proxyRequest"].get("idempotencyKey"),
            )
            if req["awaitAtTheEnd"]:
                to_await.append(awaitable)

    for awaitable in to_await:
        await awaitable



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
# pylint: disable=R0917
# pylint: disable=R0913
# pylint: disable=C0301
# pylint: disable=R0903
# pylint: disable=W0511

"""This module contains the restate context implementation based on the server"""

import asyncio
from contextlib import AsyncExitStack
import contextvars
import copy
from random import Random
from datetime import timedelta
import inspect
import functools
import logging
from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar, Union, Coroutine
import typing
import traceback
from uuid import UUID
import time

from restate.context import (
    Context,
    DurablePromise,
    AttemptFinishedEvent,
    HandlerType,
    ObjectContext,
    Request,
    RestateDurableCallFuture,
    RestateDurableFuture,
    RunAction,
    SendHandle,
    RestateDurableSleepFuture,
    RunOptions,
    P,
)
from restate.exceptions import TerminalError, SdkInternalBaseException, SdkInternalException, SuspendedException
from restate.handler import Handler, handler_from_callable, invoke_handler
from restate.serde import BytesSerde, DefaultSerde, Serde
from restate.server_types import ReceiveChannel, Send
from restate.types import extract_core_type
from restate.vm import Failure, Invocation, NotReady, VMWrapper, RunRetryConfig, Suspended  # pylint: disable=line-too-long
from restate.vm import (
    DoProgressAnyCompleted,
    DoProgressCancelSignalReceived,
    DoProgressReadFromInput,
    DoProgressExecuteRun,
    DoWaitPendingRun,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")
I = TypeVar("I")
O = TypeVar("O")


class DisconnectedException(Exception):
    """
    This exception is raised when the connection to the restate server is lost.
    This can be due to a network error or a long inactivity timeout.
    The restate-server will automatically retry the attempt.
    """

    def __init__(self) -> None:
        super().__init__("Disconnected. The connection to the restate server was lost. Restate will retry the attempt.")


class ServerTeardownEvent(AttemptFinishedEvent):
    """
    This class implements the teardown event for the server.
    """

    def __init__(self, event: asyncio.Event) -> None:
        super().__init__()
        self.event = event

    def is_set(self):
        return self.event.is_set()

    async def wait(self):
        """Wait for the event to be set."""
        await self.event.wait()


class LazyFuture(typing.Generic[T]):
    """
    Creates a task lazily, and allows multiple awaiters to the same coroutine.
    The async_def will be executed at most 1 times. (0 if __await__ or get() not called)
    """

    __slots__ = ["async_def", "task"]

    def __init__(self, async_def: Callable[[], typing.Coroutine[Any, Any, T]]) -> None:
        assert async_def is not None
        self.async_def = async_def
        self.task: asyncio.Task[T] | None = None

    def done(self):
        """
        check if completed
        """
        return self.task is not None and self.task.done()

    async def get(self) -> T:
        """Get the value of the future."""
        if self.task is None:
            self.task = asyncio.create_task(self.async_def())
        return await self.task

    def __await__(self):
        return self.get().__await__()


class ServerDurableFuture(RestateDurableFuture[T]):
    """This class implements a durable future API"""

    def __init__(self, context: "ServerInvocationContext", handle: int, async_def) -> None:
        super().__init__()
        self.context = context
        self.handle = handle
        self.future = LazyFuture(async_def)

    def is_completed(self):
        """
        A future is completed, either it was physically completed and its value has been collected.
        OR it might not yet physically completed (i.e. the async_def didn't finish yet) BUT our VM
        already has a completion value for it.
        """
        return self.future.done() or self.context.vm.is_completed(self.handle)

    def __await__(self):
        return self.future.__await__()


class ServerDurableSleepFuture(RestateDurableSleepFuture, ServerDurableFuture[None]):
    """This class implements a durable sleep future API"""

    def __await__(self) -> typing.Generator[Any, Any, None]:
        return self.future.__await__()


class ServerCallDurableFuture(RestateDurableCallFuture[T], ServerDurableFuture[T]):
    """This class implements a durable future but for calls"""

    def __init__(
        self, context: "ServerInvocationContext", result_handle: int, result_async_def, invocation_id_async_def
    ) -> None:
        super().__init__(context, result_handle, result_async_def)
        self.invocation_id_future = LazyFuture(invocation_id_async_def)

    async def invocation_id(self) -> str:
        """Get the invocation id."""
        return await self.invocation_id_future.get()

    async def cancel_invocation(self) -> None:
        """
        Cancels the invocation.

        Just a utility shortcut to:
        .. code-block:: python

           ctx.cancel_invocation(await f.invocation_id())
        """
        inv = await self.invocation_id()
        self.context.cancel_invocation(inv)


class ServerSendHandle(SendHandle):
    """This class implements the send API"""

    def __init__(self, context: "ServerInvocationContext", handle: int) -> None:
        super().__init__()
        self.context = context

        async def coro() -> str:
            if not context.vm.is_completed(handle):
                await context.create_poll_or_cancel_coroutine([handle])
            invocation_id = await context.must_take_notification(handle)
            return typing.cast(str, invocation_id)

        self.future = LazyFuture(coro)

    async def invocation_id(self) -> str:
        """Get the invocation id."""
        return await self.future

    async def cancel_invocation(self) -> None:
        """Cancel the invocation."""
        invocation_id = await self.invocation_id()
        self.context.cancel_invocation(invocation_id)


async def async_value(n: Callable[[], T]) -> T:
    """convert a simple value to a coroutine."""
    return n()


class ServerDurablePromise(DurablePromise):
    """This class implements a durable promise API"""

    def __init__(self, server_context: "ServerInvocationContext", name, serde) -> None:
        super().__init__(name=name, serde=DefaultSerde() if serde is None else serde)
        self.server_context = server_context

    def value(self) -> RestateDurableFuture[Any]:
        handle = self.server_context.vm.sys_get_promise(self.name)
        update_restate_context_is_replaying(self.server_context.vm)
        return self.server_context.create_future(handle, self.serde)

    def resolve(self, value: Any) -> Awaitable[None]:
        vm: VMWrapper = self.server_context.vm
        assert self.serde is not None
        value_buffer = self.serde.serialize(value)
        handle = vm.sys_complete_promise_success(self.name, value_buffer)
        update_restate_context_is_replaying(self.server_context.vm)

        async def await_point():
            if not self.server_context.vm.is_completed(handle):
                await self.server_context.create_poll_or_cancel_coroutine([handle])
            await self.server_context.must_take_notification(handle)

        return ServerDurableFuture(self.server_context, handle, await_point)

    def reject(self, message: str, code: int = 500) -> Awaitable[None]:
        vm: VMWrapper = self.server_context.vm
        py_failure = Failure(code=code, message=message)
        handle = vm.sys_complete_promise_failure(self.name, py_failure)
        update_restate_context_is_replaying(self.server_context.vm)

        async def await_point():
            if not self.server_context.vm.is_completed(handle):
                await self.server_context.create_poll_or_cancel_coroutine([handle])
            await self.server_context.must_take_notification(handle)

        return ServerDurableFuture(self.server_context, handle, await_point)

    def peek(self) -> Awaitable[Any | None]:
        vm: VMWrapper = self.server_context.vm
        handle = vm.sys_peek_promise(self.name)
        update_restate_context_is_replaying(self.server_context.vm)
        serde = self.serde
        assert serde is not None
        return self.server_context.create_future(handle, serde)

    def __await__(self):
        return self.value().__await__()


# disable too many public method
# pylint: disable=R0904


class Tasks:
    """
    This class implements a list of tasks.
    """

    def __init__(self) -> None:
        self.tasks: set[asyncio.Future] = set()

    def add(self, task: asyncio.Future):
        """Add a task to the list."""
        self.tasks.add(task)

        def safe_remove(_):
            """Remove the task from the list."""
            try:
                self.tasks.remove(task)
            except KeyError:
                pass
            try:
                _ = task.result()
            # pylint: disable=W0718
            except Exception:
                pass
            except asyncio.CancelledError:
                pass

        task.add_done_callback(safe_remove)

    def cancel(self):
        """Cancel all tasks in the list."""
        to_cancel = list(self.tasks)
        self.tasks.clear()
        for task in to_cancel:
            task.cancel()


restate_context_is_replaying = contextvars.ContextVar("restate_context_is_replaying", default=False)


def update_restate_context_is_replaying(vm: VMWrapper):
    """Update the context var 'restate_context_is_replaying'. This should be called after each vm.sys_*"""
    restate_context_is_replaying.set(vm.is_replaying())


_restate_context_var = contextvars.ContextVar[Context]("restate_context")


def current_context() -> Context | None:
    """Get the current context."""
    return _restate_context_var.get()


# pylint: disable=R0902
class ServerInvocationContext(ObjectContext):
    """This class implements the context for the restate framework based on the server."""

    def __init__(
        self,
        vm: VMWrapper,
        handler: Handler[I, O],
        invocation: Invocation,
        attempt_headers: Dict[str, str],
        send: Send,
        receive: ReceiveChannel,
    ) -> None:
        super().__init__()
        self.vm = vm
        self.handler = handler
        self.invocation = invocation
        self.attempt_headers = attempt_headers
        self.send = send
        self.random_instance = Random(invocation.random_seed)
        self.receive = receive
        self.run_coros_to_execute: dict[int, Callable[[], Awaitable[None]]] = {}
        self.request_finished_event = asyncio.Event()
        self.tasks = Tasks()

    async def enter(self):
        """Invoke the user code."""
        update_restate_context_is_replaying(self.vm)
        token = _restate_context_var.set(self)
        try:
            in_buffer = self.invocation.input_buffer
            async with AsyncExitStack() as stack:
                for manager in self.handler.context_managers or []:
                    await stack.enter_async_context(manager())
                out_buffer = await invoke_handler(handler=self.handler, ctx=self, in_buffer=in_buffer)
            restate_context_is_replaying.set(False)
            self.vm.sys_write_output_success(bytes(out_buffer))
            self.vm.sys_end()
        except TerminalError as t:
            failure = Failure(code=t.status_code, message=t.message)
            restate_context_is_replaying.set(False)
            self.vm.sys_write_output_failure(failure)
            self.vm.sys_end()
        # pylint: disable=W0718
        except asyncio.CancelledError:
            pass
        except SdkInternalBaseException:
            pass
        except DisconnectedException:
            raise
        except Exception as e:
            # check the immediate cause for SdkInternalBaseException or TerminalError that
            # got accidentally re-raised by user code as part of another exception.
            cause: BaseException | None = e
            while cause is not None:
                if isinstance(cause, TerminalError):
                    failure = Failure(code=cause.status_code, message=cause.message)
                    restate_context_is_replaying.set(False)
                    self.vm.sys_write_output_failure(failure)
                    self.vm.sys_end()
                    break
                elif isinstance(cause, SdkInternalBaseException):
                    break
                cause = cause.__cause__
            else:
                # nothing interesting found, treat as unexpected exception
                stacktrace = "\n".join(traceback.format_exception(e))
                self.vm.notify_error(repr(e), stacktrace)
                raise
        finally:
            _restate_context_var.reset(token)

    async def leave(self):
        """Leave the context."""
        while True:
            chunk = self.vm.take_output()
            if chunk is None:
                break
            await self.send(
                {
                    "type": "http.response.body",
                    "body": chunk,
                    "more_body": True,
                }
            )
        # ========================================
        # End the connection
        # ========================================
        # The following sequence of events is expected during a teardown:
        #
        # {'type': 'http.request', 'body': b'', 'more_body': True}
        # {'type': 'http.request', 'body': b'', 'more_body': False}
        # {'type': 'http.disconnect'}
        await self.receive.block_until_http_input_closed()
        # finally, we close our side
        # it is important to do it, after the other side has closed his side,
        # because some asgi servers (like hypercorn) will remove the stream
        # as soon as they see a close event (in asgi terms more_body=False)
        await self.send(
            {
                "type": "http.response.body",
                "body": b"",
                "more_body": False,
            }
        )
        # notify to any holder of the abort signal that we are done

    def on_attempt_finished(self):
        """Notify the attempt finished event."""
        restate_context_is_replaying.set(False)
        self.request_finished_event.set()
        try:
            self.tasks.cancel()
        except asyncio.CancelledError:
            # ignore the cancelled error
            pass

    async def take_and_send_output(self):
        """Take output from state machine and send it"""
        output = self.vm.take_output()
        if output:
            await self.send(
                {
                    "type": "http.response.body",
                    "body": bytes(output),
                    "more_body": True,
                }
            )

    async def must_take_notification(self, handle):
        """Take notification, which must be present"""
        res = self.vm.take_notification(handle)
        if isinstance(res, Exception):
            # We might need to write out something at this point.
            await self.take_and_send_output()
            logger.exception("Exception in take_notification", exc_info=res)
            raise SdkInternalException() from res
        if isinstance(res, Suspended):
            # We might need to write out something at this point.
            await self.take_and_send_output()
            raise SuspendedException()
        if isinstance(res, NotReady):
            raise ValueError(f"Unexpected value error: {handle}")
        if res is None:
            return None
        if isinstance(res, Failure):
            raise TerminalError(res.message, res.code)
        return res

    async def create_poll_or_cancel_coroutine(self, handles: typing.List[int]) -> None:
        """Create a coroutine to poll the handle."""
        while True:
            await self.take_and_send_output()
            do_progress_response = self.vm.do_progress(handles)
            if isinstance(do_progress_response, BaseException):
                logger.exception("Exception in do_progress", exc_info=do_progress_response)
                raise SdkInternalException() from do_progress_response
            if isinstance(do_progress_response, Suspended):
                raise SuspendedException()
            if isinstance(do_progress_response, DoProgressAnyCompleted):
                # One of the handles completed
                return
            if isinstance(do_progress_response, DoProgressCancelSignalReceived):
                raise TerminalError("cancelled", 409)
            if isinstance(do_progress_response, DoProgressExecuteRun):
                fn = self.run_coros_to_execute[do_progress_response.handle]
                del self.run_coros_to_execute[do_progress_response.handle]
                assert fn is not None

                async def wrapper(f):
                    try:
                        await f()
                    finally:
                        await self.receive.enqueue_restate_event({"type": "restate.run_completed", "data": None})

                task = asyncio.create_task(wrapper(fn))
                self.tasks.add(task)
                continue
            if isinstance(do_progress_response, (DoWaitPendingRun, DoProgressReadFromInput)):
                chunk = await self.receive()
                if chunk.get("type") == "restate.run_completed":
                    continue
                if chunk.get("type") == "http.disconnect":
                    raise DisconnectedException()
                if chunk.get("body", None) is not None:
                    body = chunk.get("body")
                    assert isinstance(body, bytes)
                    self.vm.notify_input(body)
                if not chunk.get("more_body", False):
                    self.vm.notify_input_closed()

    def _create_fetch_result_coroutine(self, handle: int, serde: Serde[T] | None = None):
        """Create a coroutine that fetches a result from a notification handle."""

        async def fetch_result():
            if not self.vm.is_completed(handle):
                await self.create_poll_or_cancel_coroutine([handle])
            res = await self.must_take_notification(handle)
            if res is None or serde is None:
                return res
            if isinstance(res, bytes):
                return serde.deserialize(res)
            return res

        return fetch_result

    def create_future(self, handle: int, serde: Serde[T] | None = None) -> ServerDurableFuture[T]:
        """Create a durable future for handling asynchronous state operations."""
        return ServerDurableFuture(self, handle, self._create_fetch_result_coroutine(handle, serde))

    def create_sleep_future(self, handle: int) -> ServerDurableSleepFuture:
        """Create a durable sleep future."""

        async def transform():
            if not self.vm.is_completed(handle):
                await self.create_poll_or_cancel_coroutine([handle])
            await self.must_take_notification(handle)

        return ServerDurableSleepFuture(self, handle, transform)

    def create_call_future(
        self, handle: int, invocation_id_handle: int, serde: Serde[T] | None = None
    ) -> ServerCallDurableFuture[T]:
        """Create a durable future."""

        async def inv_id_factory():
            if not self.vm.is_completed(invocation_id_handle):
                await self.create_poll_or_cancel_coroutine([invocation_id_handle])
            return await self.must_take_notification(invocation_id_handle)

        return ServerCallDurableFuture(self, handle, self._create_fetch_result_coroutine(handle, serde), inv_id_factory)

    def get(
        self, name: str, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> Awaitable[Optional[T]]:
        handle = self.vm.sys_get_state(name)
        update_restate_context_is_replaying(self.vm)
        if isinstance(serde, DefaultSerde):
            serde = serde.with_maybe_type(type_hint)
        return self.create_future(handle, serde)  # type: ignore

    def state_keys(self) -> Awaitable[List[str]]:
        handle = self.vm.sys_get_state_keys()
        update_restate_context_is_replaying(self.vm)
        return self.create_future(handle)

    def set(self, name: str, value: T, serde: Serde[T] = DefaultSerde()) -> None:
        """Set the value associated with the given name."""
        if isinstance(serde, DefaultSerde):
            serde = serde.with_maybe_type(type(value))
        buffer = serde.serialize(value)
        self.vm.sys_set_state(name, bytes(buffer))
        update_restate_context_is_replaying(self.vm)

    def clear(self, name: str) -> None:
        self.vm.sys_clear_state(name)
        update_restate_context_is_replaying(self.vm)

    def clear_all(self) -> None:
        self.vm.sys_clear_all_state()
        update_restate_context_is_replaying(self.vm)

    def request(self) -> Request:
        return Request(
            id=self.invocation.invocation_id,
            headers=dict(self.invocation.headers),
            attempt_headers=self.attempt_headers,
            body=self.invocation.input_buffer,
            attempt_finished_event=ServerTeardownEvent(self.request_finished_event),
        )

    def random(self) -> Random:
        return self.random_instance

    def uuid(self) -> UUID:
        return UUID(int=self.random_instance.getrandbits(128), version=4)

    def time(self) -> RestateDurableFuture[float]:
        # pylint: disable=W0108
        return self.run_typed("timestamp", lambda: time.time())

    # pylint: disable=R0914
    async def create_run_coroutine(
        self,
        handle: int,
        action: RunAction[T],
        serde: Serde[T],
        max_attempts: Optional[int] = None,
        max_duration: Optional[timedelta] = None,
        initial_retry_interval: Optional[timedelta] = None,
        max_retry_interval: Optional[timedelta] = None,
        retry_interval_factor: Optional[float] = None,
    ) -> None:
        """Create a coroutine to poll the handle."""
        start = time.time()
        try:
            if inspect.iscoroutinefunction(action):
                action_result: T = await action()  # type: ignore
            else:
                loop = asyncio.get_running_loop()
                ctx = contextvars.copy_context()
                func_call = functools.partial(ctx.run, action)
                action_result_future = loop.run_in_executor(None, func_call)
                self.tasks.add(action_result_future)
                action_result = typing.cast(T, await action_result_future)

            buffer = serde.serialize(action_result)
            self.vm.propose_run_completion_success(handle, buffer)
        except TerminalError as t:
            failure = Failure(code=t.status_code, message=t.message)
            self.vm.propose_run_completion_failure(handle, failure)
        except asyncio.CancelledError as e:
            raise e from None
        except SdkInternalBaseException as e:
            raise e from None
        # pylint: disable=W0718
        except Exception as e:
            end = time.time()
            attempt_duration = int((end - start) * 1000)
            failure = Failure(code=500, message=str(e))
            max_duration_ms = None if max_duration is None else int(max_duration.total_seconds() * 1000)
            initial_retry_interval_ms = (
                None if initial_retry_interval is None else int(initial_retry_interval.total_seconds() * 1000)
            )
            max_retry_interval_ms = (
                None if max_retry_interval is None else int(max_retry_interval.total_seconds() * 1000)
            )
            config = RunRetryConfig(
                max_attempts=max_attempts,
                max_duration=max_duration_ms,
                initial_interval=initial_retry_interval_ms,
                max_interval=max_retry_interval_ms,
                interval_factor=retry_interval_factor,
            )
            self.vm.propose_run_completion_transient(
                handle, failure=failure, attempt_duration_ms=attempt_duration, config=config
            )

    # pylint: disable=W0236
    # pylint: disable=R0914
    def run(
        self,
        name: str,
        action: RunAction[T],
        serde: Serde[T] = DefaultSerde(),
        max_attempts: Optional[int] = None,
        max_retry_duration: Optional[timedelta] = None,
        type_hint: Optional[typing.Type[T]] = None,
        args: Optional[typing.Tuple[Any, ...]] = None,
    ) -> RestateDurableFuture[T]:
        if isinstance(serde, DefaultSerde):
            if type_hint is None:
                signature = inspect.signature(action, eval_str=True)
                type_hint = signature.return_annotation
            serde = serde.with_maybe_type(type_hint)

        handle = self.vm.sys_run(name)
        update_restate_context_is_replaying(self.vm)

        if args is not None:
            noargs_action = typing.cast(RunAction[T], functools.partial(action, *args))
        else:
            noargs_action = action
        self.run_coros_to_execute[handle] = lambda: self.create_run_coroutine(
            handle, noargs_action, serde, max_attempts, max_retry_duration, None, None, None
        )
        return self.create_future(handle, serde)  # type: ignore

    def run_typed(
        self,
        name: str,
        action: Union[Callable[P, T], Callable[P, Coroutine[Any, Any, T]]],
        options: RunOptions[T] = RunOptions(),
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RestateDurableFuture[T]:
        if isinstance(options.serde, DefaultSerde):
            # Copy options, as we're going to mutate type_hint and serde
            options = copy.copy(options)
            if options.type_hint is None:
                signature = inspect.signature(action, eval_str=True)
                options.type_hint = signature.return_annotation
            core_type_kind, core_type = extract_core_type(options.type_hint)
            if core_type_kind == "simple" or core_type_kind == "optional":
                # use core type as it is more specific. E.g. Optional[T] -> T
                options.type_hint = core_type
            options.serde = typing.cast(DefaultSerde, options.serde).with_maybe_type(options.type_hint)
        handle = self.vm.sys_run(name)
        update_restate_context_is_replaying(self.vm)

        func = typing.cast(RunAction[T], functools.partial(action, *args, **kwargs))
        self.run_coros_to_execute[handle] = lambda: self.create_run_coroutine(
            handle,
            func,
            options.serde,
            options.max_attempts,
            options.max_duration,
            options.initial_retry_interval,
            options.max_retry_interval,
            options.retry_interval_factor,
        )
        return self.create_future(handle, options.serde)

    def sleep(self, delta: timedelta, name: Optional[str] = None) -> RestateDurableSleepFuture:
        # convert timedelta to milliseconds
        millis = int(delta.total_seconds() * 1000)
        handle = self.vm.sys_sleep(millis, name)
        update_restate_context_is_replaying(self.vm)
        return self.create_sleep_future(handle)  # type: ignore

    def do_call(
        self,
        tpe: HandlerType[I, O],
        parameter: I,
        key: Optional[str] = None,
        send_delay: Optional[timedelta] = None,
        send: bool = False,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O] | SendHandle:
        """Make an RPC call to the given handler"""
        target_handler = handler_from_callable(tpe)
        service = target_handler.service_tag.name
        handler = target_handler.name
        input_serde = target_handler.handler_io.input_serde
        output_serde = target_handler.handler_io.output_serde
        return self.do_raw_call(
            service, handler, parameter, input_serde, output_serde, key, send_delay, send, idempotency_key, headers
        )

    def do_raw_call(
        self,
        service: str,
        handler: str,
        input_param: I,
        input_serde: Serde[I],
        output_serde: Serde[O],
        key: Optional[str] = None,
        send_delay: Optional[timedelta] = None,
        send: bool = False,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O] | SendHandle:
        """Make an RPC call to the given handler"""
        parameter = input_serde.serialize(input_param)
        if headers is not None:
            headers_kvs = list(headers.items())
        else:
            headers_kvs = []
        if send_delay:
            ms = int(send_delay.total_seconds() * 1000)
            send_handle = self.vm.sys_send(
                service, handler, parameter, key, delay=ms, idempotency_key=idempotency_key, headers=headers_kvs
            )
            update_restate_context_is_replaying(self.vm)
            return ServerSendHandle(self, send_handle)
        if send:
            send_handle = self.vm.sys_send(
                service, handler, parameter, key, idempotency_key=idempotency_key, headers=headers_kvs
            )
            update_restate_context_is_replaying(self.vm)
            return ServerSendHandle(self, send_handle)

        handle = self.vm.sys_call(
            service=service,
            handler=handler,
            parameter=parameter,
            key=key,
            idempotency_key=idempotency_key,
            headers=headers_kvs,
        )
        update_restate_context_is_replaying(self.vm)

        return self.create_call_future(
            handle=handle.result_handle, invocation_id_handle=handle.invocation_id_handle, serde=output_serde
        )

    def service_call(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O]:
        coro = self.do_call(tpe, arg, idempotency_key=idempotency_key, headers=headers)
        assert not isinstance(coro, SendHandle)
        return coro

    def service_send(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        send_delay: timedelta | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        send = self.do_call(
            tpe=tpe, parameter=arg, send_delay=send_delay, send=True, idempotency_key=idempotency_key, headers=headers
        )
        assert isinstance(send, SendHandle)
        return send

    def object_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O]:
        coro = self.do_call(tpe, arg, key, idempotency_key=idempotency_key, headers=headers)
        assert not isinstance(coro, SendHandle)
        return coro

    def object_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: timedelta | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        send = self.do_call(
            tpe=tpe,
            key=key,
            parameter=arg,
            send_delay=send_delay,
            send=True,
            idempotency_key=idempotency_key,
            headers=headers,
        )
        assert isinstance(send, SendHandle)
        return send

    def workflow_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O]:
        return self.object_call(tpe, key, arg, idempotency_key=idempotency_key, headers=headers)

    def workflow_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: timedelta | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        send = self.object_send(tpe, key, arg, send_delay, idempotency_key=idempotency_key, headers=headers)
        assert isinstance(send, SendHandle)
        return send

    def generic_call(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: str | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[bytes]:
        serde = BytesSerde()
        call_handle = self.do_raw_call(
            service=service,
            handler=handler,
            input_param=arg,
            input_serde=serde,
            output_serde=serde,
            key=key,
            idempotency_key=idempotency_key,
            headers=headers,
        )
        assert not isinstance(call_handle, SendHandle)
        return call_handle

    def generic_send(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: str | None = None,
        send_delay: timedelta | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        serde = BytesSerde()
        send_handle = self.do_raw_call(
            service=service,
            handler=handler,
            input_param=arg,
            input_serde=serde,
            output_serde=serde,
            key=key,
            send_delay=send_delay,
            send=True,
            idempotency_key=idempotency_key,
            headers=headers,
        )
        assert isinstance(send_handle, SendHandle)
        return send_handle

    def awakeable(
        self, serde: Serde[I] = DefaultSerde(), type_hint: Optional[typing.Type[I]] = None
    ) -> typing.Tuple[str, RestateDurableFuture[Any]]:
        if isinstance(serde, DefaultSerde):
            serde = serde.with_maybe_type(type_hint)
        name, handle = self.vm.sys_awakeable()
        update_restate_context_is_replaying(self.vm)
        return name, self.create_future(handle, serde)

    def resolve_awakeable(self, name: str, value: I, serde: Serde[I] = DefaultSerde()) -> None:
        if isinstance(serde, DefaultSerde):
            serde = serde.with_maybe_type(type(value))
        buf = serde.serialize(value)
        self.vm.sys_resolve_awakeable(name, buf)
        update_restate_context_is_replaying(self.vm)

    def reject_awakeable(self, name: str, failure_message: str, failure_code: int = 500) -> None:
        self.vm.sys_reject_awakeable(name, Failure(code=failure_code, message=failure_message))
        update_restate_context_is_replaying(self.vm)

    def promise(
        self, name: str, serde: typing.Optional[Serde[T]] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> DurablePromise[T]:
        """Create a durable promise."""
        if isinstance(serde, DefaultSerde):
            serde = serde.with_maybe_type(type_hint)
        return ServerDurablePromise(self, name, serde)

    def key(self) -> str:
        return self.invocation.key

    def cancel_invocation(self, invocation_id: str):
        """cancel an existing invocation by id."""
        if invocation_id is None:
            raise ValueError("invocation_id cannot be None")
        self.vm.sys_cancel(invocation_id)
        update_restate_context_is_replaying(self.vm)

    def attach_invocation(
        self, invocation_id: str, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> RestateDurableFuture[T]:
        if invocation_id is None:
            raise ValueError("invocation_id cannot be None")
        if isinstance(serde, DefaultSerde):
            serde = serde.with_maybe_type(type_hint)
        handle = self.vm.attach_invocation(invocation_id)
        update_restate_context_is_replaying(self.vm)
        return self.create_future(handle, serde)



from restate.serde import BytesSerde


def test_bytes_serde():
    s = BytesSerde()
    assert bytes(range(20)) == s.serialize(bytes(range(20)))


def extract_core_type_optional():
    from restate.types import extract_core_type

    from typing import Optional, Union

    kind, tpe = extract_core_type(Optional[int])
    assert kind == "optional"
    assert tpe is int

    kind, tpe = extract_core_type(Union[int, None])
    assert kind == "optional"
    assert tpe is int

    kind, tpe = extract_core_type(str | None)
    assert kind == "optional"
    assert tpe is str

    kind, tpe = extract_core_type(None | str)
    assert kind == "optional"
    assert tpe is str

    for t in [int, str, bytes, dict, list, None]:
        kind, tpe = extract_core_type(t)
        assert kind == "simple"
        assert tpe is t



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

import uuid
import restate
from restate import (
    Context,
    Service,
    HarnessEnvironment,
    VirtualObject,
    ObjectContext,
    ObjectSharedContext,
    Workflow,
    WorkflowContext,
    getLogger,
    WorkflowSharedContext,
)
import pytest
import asyncio

# ----- Asyncio fixtures


@pytest.fixture(scope="session")
def anyio_backend():
    return "asyncio"


pytestmark = [
    pytest.mark.anyio,
]

# -------- Restate services and restate fixture

greeter = Service("greeter")


@greeter.handler()
async def greet(ctx: Context, name: str) -> str:
    return f"Hello {name}!"


counter = VirtualObject("counter")


@counter.handler()
async def increment(ctx: ObjectContext, value: int) -> int:
    n = await ctx.get("counter", type_hint=int) or 0
    n += value
    ctx.set("counter", n)
    return n


@counter.handler(kind="shared")
async def count(ctx: ObjectSharedContext) -> int:
    return await ctx.get("counter") or 0


payment = Workflow("payment")
payment_logger = getLogger("payment")


@payment.main()
async def pay(ctx: WorkflowContext):
    ctx.set("status", "verifying payment")

    def payment_gateway():
        payment_logger.info("Doing payment work")

    await ctx.run_typed("payment", payment_gateway)

    ctx.set("status", "waiting for the payment provider to approve")

    # Wait for the payment to be verified
    result = await ctx.promise("verify.payment", type_hint=str).value()
    return f"Verified {result}!"


@payment.handler()
async def payment_verified(ctx: WorkflowSharedContext, result: str):
    promise = ctx.promise("verify.payment", type_hint=str)
    await promise.resolve(result)


@pytest.fixture(scope="session")
async def restate_test_harness():
    async with restate.create_test_harness(
        restate.app([greeter, counter, payment]), restate_image="ghcr.io/restatedev/restate:latest"
    ) as harness:
        yield harness


# ----- Tests


async def test_greeter(restate_test_harness: HarnessEnvironment):
    greeting = await restate_test_harness.client.service_call(greet, arg="Pippo")

    assert greeting == "Hello Pippo!"


async def test_counter(restate_test_harness: HarnessEnvironment):
    random_key = str(uuid.uuid4())
    initial_count = await restate_test_harness.client.object_call(count, key=random_key, arg=None)
    await restate_test_harness.client.object_call(increment, key=random_key, arg=1)
    new_count = await restate_test_harness.client.object_call(count, key=random_key, arg=None)

    assert new_count == initial_count + 1


async def test_idempotency_key(restate_test_harness: HarnessEnvironment):
    random_key = str(uuid.uuid4())
    initial_count = await restate_test_harness.client.object_call(count, key=random_key, arg=None)
    await restate_test_harness.client.object_call(increment, key=random_key, arg=1, idempotency_key=random_key)
    await restate_test_harness.client.object_call(increment, key=random_key, arg=1, idempotency_key=random_key)
    new_count = await restate_test_harness.client.object_call(count, key=random_key, arg=None)

    assert new_count == initial_count + 1


async def test_workflow(restate_test_harness: HarnessEnvironment):
    random_key = str(uuid.uuid4())
    call_task = asyncio.create_task(restate_test_harness.client.workflow_call(pay, key=random_key, arg=None))

    await restate_test_harness.client.workflow_call(payment_verified, key=random_key, arg="Done")

    assert await call_task == "Verified Done!"


async def test_send(restate_test_harness: HarnessEnvironment):
    invocation_handle = await restate_test_harness.client.service_send(greet, arg="Pippo")

    assert invocation_handle.status_code == 200
    assert len(invocation_handle.invocation_id) > 0



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
ADK plugin implementation for restate.
"""

import asyncio
import restate

from datetime import timedelta
from typing import Optional, Any, cast

from google.genai import types

from google.adk.agents import BaseAgent, LlmAgent
from google.adk.agents.callback_context import CallbackContext
from google.adk.plugins import BasePlugin
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext
from google.adk.models.llm_request import LlmRequest
from google.adk.models.llm_response import LlmResponse
from google.adk.models import LLMRegistry
from google.adk.models.base_llm import BaseLlm
from google.adk.flows.llm_flows.functions import generate_client_function_call_id


from restate.extensions import current_context

from .session import flush_session_state


class RestatePlugin(BasePlugin):
    """A plugin to integrate Restate with the ADK framework."""

    _models: dict[str, BaseLlm]
    _locks: dict[str, asyncio.Lock]

    def __init__(self, *, max_model_call_retries: int = 10):
        super().__init__(name="restate_plugin")
        self._models = {}
        self._locks = {}
        self._max_model_call_retries = max_model_call_retries

    async def before_agent_callback(
        self, *, agent: BaseAgent, callback_context: CallbackContext
    ) -> Optional[types.Content]:
        if not isinstance(agent, LlmAgent):
            raise restate.TerminalError("RestatePlugin only supports LlmAgent agents.")
        ctx = current_context()  # Ensure we have a Restate context
        if ctx is None:
            raise restate.TerminalError(
                """No Restate context found for RestatePlugin.
            Ensure that the agent is invoked within a restate handler and,
            using a ```with restate_overrides(ctx):``` block. around your agent use."""
            )
        model = agent.model if isinstance(agent.model, BaseLlm) else LLMRegistry.new_llm(agent.model)
        self._models[callback_context.invocation_id] = model
        self._locks[callback_context.invocation_id] = asyncio.Lock()

        id = callback_context.invocation_id
        event = ctx.request().attempt_finished_event

        async def release_task():
            """make sure to release resources when the agent finishes"""
            try:
                await event.wait()
            finally:
                self._models.pop(id, None)
                self._locks.pop(id, None)

        _ = asyncio.create_task(release_task())
        return None

    async def after_agent_callback(
        self, *, agent: BaseAgent, callback_context: CallbackContext
    ) -> Optional[types.Content]:
        self._models.pop(callback_context.invocation_id, None)
        self._locks.pop(callback_context.invocation_id, None)

        ctx = cast(restate.ObjectContext, current_context())
        await flush_session_state(ctx, callback_context.session)

        return None

    async def before_model_callback(
        self, *, callback_context: CallbackContext, llm_request: LlmRequest
    ) -> Optional[LlmResponse]:
        model = self._models[callback_context.invocation_id]
        ctx = current_context()
        if ctx is None:
            raise RuntimeError(
                "No Restate context found, the restate plugin must be used from within a restate handler."
            )
        response = await _generate_content_async(ctx, self._max_model_call_retries, model, llm_request)
        return response

    async def before_tool_callback(
        self,
        *,
        tool: BaseTool,
        tool_args: dict[str, Any],
        tool_context: ToolContext,
    ) -> Optional[dict]:
        tool_context.session.state["restate_context"] = current_context()
        lock = self._locks[tool_context.invocation_id]
        await lock.acquire()
        # TODO: if we want we can also automatically wrap tools with ctx.run_typed here
        return None

    async def after_tool_callback(
        self,
        *,
        tool: BaseTool,
        tool_args: dict[str, Any],
        tool_context: ToolContext,
        result: dict,
    ) -> Optional[dict]:
        lock = self._locks[tool_context.invocation_id]
        lock.release()
        tool_context.session.state.pop("restate_context", None)
        return None

    async def on_tool_error_callback(
        self,
        *,
        tool: BaseTool,
        tool_args: dict[str, Any],
        tool_context: ToolContext,
        error: Exception,
    ) -> Optional[dict]:
        lock = self._locks[tool_context.invocation_id]
        lock.release()
        tool_context.session.state.pop("restate_context", None)
        return None

    async def close(self):
        self._models.clear()
        self._locks.clear()


def _generate_client_function_call_id(s: LlmResponse) -> None:
    """Generate client function call IDs for function calls in the LlmResponse.
    It is important for the function call IDs to be stable across retries, as they
    are used to correlate function call results with their invocations.
    """
    if s.content and s.content.parts:
        for part in s.content.parts:
            if part.function_call:
                if not part.function_call.id:
                    id = generate_client_function_call_id()
                    part.function_call.id = id


async def _generate_content_async(
    ctx: restate.Context, max_attempts: int, model: BaseLlm, llm_request: LlmRequest
) -> LlmResponse:
    """Generate content using Restate's context."""

    async def call_llm() -> LlmResponse:
        a_gen = model.generate_content_async(llm_request, stream=False)
        try:
            result = await anext(a_gen)
            _generate_client_function_call_id(result)
            return result
        finally:
            await a_gen.aclose()

    return await ctx.run_typed(
        "call LLM",
        call_llm,
        restate.RunOptions(max_attempts=max_attempts, initial_retry_interval=timedelta(seconds=1)),
    )



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains the restate exceptions"""

# pylint: disable=C0301


class TerminalError(Exception):
    """This exception is thrown to indicate that Restate should not retry."""

    def __init__(self, message: str, status_code: int = 500) -> None:
        super().__init__(message)
        self.message = message
        self.status_code = status_code


class SdkInternalBaseException(BaseException):
    """This exception is internal, and you should not catch it.
    If you need to distinguish with other exceptions, use is_internal_exception."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message
            + """
This exception is safe to ignore. If you see it, you might be using a try/catch all statement.

Don't do:
try:
  # Code
except:
  # This catches all exceptions, including the SdkInternalBaseException!

Do instead:
try:
  # Code
except TerminalError:
  # In Restate handlers you typically want to catch TerminalError only

Or remove the try/except altogether if you don't need it.
For further info on error handling, refer to https://docs.restate.dev/develop/python/error-handling
"""
        )


class SuspendedException(SdkInternalBaseException):
    """This exception is raised to indicate that the execution is suspended"""

    def __init__(self) -> None:
        super().__init__("Invocation got suspended, Restate will resume this invocation when progress can be made.")


class SdkInternalException(SdkInternalBaseException):
    """This exception is raised to indicate that the execution raised a retryable error."""

    def __init__(self) -> None:
        super().__init__(
            "Invocation attempt raised a retryable error.\n"
            "Restate will retry executing this invocation from the point where it left off."
        )


def is_internal_exception(e) -> bool:
    """Returns true if the exception is an internal Restate exception"""
    return isinstance(e, SdkInternalBaseException)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the optional OpenAI integration for Restate.
"""

import asyncio
import dataclasses
import typing

from agents import (
    Tool,
    Usage,
    Model,
    RunContextWrapper,
    AgentsException,
    Runner as OpenAIRunner,
    RunConfig,
    TContext,
    RunResult,
    Agent,
    ModelBehaviorError,
)

from agents.models.multi_provider import MultiProvider
from agents.items import TResponseStreamEvent, TResponseOutputItem, ModelResponse
from agents.memory.session import SessionABC
from agents.items import TResponseInputItem
from typing import List, Any
from typing import AsyncIterator

from agents.tool import FunctionTool
from agents.tool_context import ToolContext
from pydantic import BaseModel
from restate.exceptions import SdkInternalBaseException
from restate.extensions import current_context

from restate import RunOptions, ObjectContext, TerminalError


# The OpenAI ModelResponse class is a dataclass with Pydantic fields.
# The Restate SDK cannot serialize this. So we turn the ModelResponse int a Pydantic model.
class RestateModelResponse(BaseModel):
    output: list[TResponseOutputItem]
    """A list of outputs (messages, tool calls, etc) generated by the model"""

    usage: Usage
    """The usage information for the response."""

    response_id: str | None
    """An ID for the response which can be used to refer to the response in subsequent calls to the
    model. Not supported by all model providers.
    If using OpenAI models via the Responses API, this is the `response_id` parameter, and it can
    be passed to `Runner.run`.
    """

    def to_input_items(self) -> list[TResponseInputItem]:
        return [it.model_dump(exclude_unset=True) for it in self.output]  # type: ignore


class DurableModelCalls(MultiProvider):
    """
    A Restate model provider that wraps the OpenAI SDK's default MultiProvider.
    """

    def __init__(self, max_retries: int | None = 3):
        super().__init__()
        self.max_retries = max_retries

    def get_model(self, model_name: str | None) -> Model:
        return RestateModelWrapper(super().get_model(model_name or None), self.max_retries)


class RestateModelWrapper(Model):
    """
    A wrapper around the OpenAI SDK's Model that persists LLM calls in the Restate journal.
    """

    def __init__(self, model: Model, max_retries: int | None = 3):
        self.model = model
        self.model_name = "RestateModelWrapper"
        self.max_retries = max_retries

    async def get_response(self, *args, **kwargs) -> ModelResponse:
        async def call_llm() -> RestateModelResponse:
            resp = await self.model.get_response(*args, **kwargs)
            # convert to pydantic model to be serializable by Restate SDK
            return RestateModelResponse(
                output=resp.output,
                usage=resp.usage,
                response_id=resp.response_id,
            )

        ctx = current_context()
        if ctx is None:
            raise RuntimeError("No current Restate context found, make sure to run inside a Restate handler")
        result = await ctx.run_typed("call LLM", call_llm, RunOptions(max_attempts=self.max_retries))
        # convert back to original ModelResponse
        return ModelResponse(
            output=result.output,
            usage=result.usage,
            response_id=result.response_id,
        )

    def stream_response(self, *args, **kwargs) -> AsyncIterator[TResponseStreamEvent]:
        raise TerminalError("Streaming is not supported in Restate. Use `get_response` instead.")


class RestateSession(SessionABC):
    """Restate session implementation following the Session protocol."""

    def _ctx(self) -> ObjectContext:
        return typing.cast(ObjectContext, current_context())

    async def get_items(self, limit: int | None = None) -> List[TResponseInputItem]:
        """Retrieve conversation history for this session."""
        current_items = await self._ctx().get("items", type_hint=List[TResponseInputItem]) or []
        if limit is not None:
            return current_items[-limit:]
        return current_items

    async def add_items(self, items: List[TResponseInputItem]) -> None:
        """Store new items for this session."""
        # Your implementation here
        current_items = await self.get_items() or []
        self._ctx().set("items", current_items + items)

    async def pop_item(self) -> TResponseInputItem | None:
        """Remove and return the most recent item from this session."""
        current_items = await self.get_items() or []
        if current_items:
            item = current_items.pop()
            self._ctx().set("items", current_items)
            return item
        return None

    async def clear_session(self) -> None:
        """Clear all items for this session."""
        self._ctx().clear("items")


class AgentsTerminalException(AgentsException, TerminalError):
    """Exception that is both an AgentsException and a restate.TerminalError."""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AgentsSuspension(AgentsException, SdkInternalBaseException):
    """Exception that is both an AgentsException and a restate SdkInternalBaseException."""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def raise_terminal_errors(context: RunContextWrapper[Any], error: Exception) -> str:
    """A custom function to provide a user-friendly error message."""
    # Raise terminal errors and cancellations
    if isinstance(error, TerminalError):
        # For the agent SDK it needs to be an AgentsException, for restate it needs to be a TerminalError
        # so we create a new exception that inherits from both
        raise AgentsTerminalException(error.message)

    if isinstance(error, ModelBehaviorError):
        return f"An error occurred while calling the tool: {str(error)}"

    raise error


def continue_on_terminal_errors(context: RunContextWrapper[Any], error: Exception) -> str:
    """A custom function to provide a user-friendly error message."""
    # Raise terminal errors and cancellations
    if isinstance(error, TerminalError):
        # For the agent SDK it needs to be an AgentsException, for restate it needs to be a TerminalError
        # so we create a new exception that inherits from both
        return f"An error occurred while running the tool: {str(error)}"

    if isinstance(error, ModelBehaviorError):
        return f"An error occurred while calling the tool: {str(error)}"

    raise error


class Runner:
    """
    A wrapper around Runner.run that automatically configures RunConfig for Restate contexts.

    This class automatically sets up the appropriate model provider (DurableModelCalls) and
    model settings, taking over any model and model_settings configuration provided in the
    original RunConfig.
    """

    @staticmethod
    async def run(
        starting_agent: Agent[TContext],
        disable_tool_autowrapping: bool = False,
        *args: typing.Any,
        run_config: RunConfig | None = None,
        **kwargs,
    ) -> RunResult:
        """
        Run an agent with automatic Restate configuration.

        Returns:
            The result from Runner.run
        """

        current_run_config = run_config or RunConfig()
        new_run_config = dataclasses.replace(
            current_run_config,
            model_provider=DurableModelCalls(),
        )
        restate_agent = sequentialize_and_wrap_tools(starting_agent, disable_tool_autowrapping)
        return await OpenAIRunner.run(restate_agent, *args, run_config=new_run_config, **kwargs)


def sequentialize_and_wrap_tools(
    agent: Agent[TContext],
    disable_tool_autowrapping: bool,
) -> Agent[TContext]:
    """
    Wrap the tools of an agent to use the Restate error handling.

    Returns:
        A new agent with wrapped tools.
    """

    # Restate does not allow parallel tool calls, so we use a lock to ensure sequential execution.
    # This lock only affects tools for this agent; handoff agents are wrapped recursively.
    sequential_tools_lock = asyncio.Lock()
    wrapped_tools: list[Tool] = []
    for tool in agent.tools:
        if isinstance(tool, FunctionTool):

            def create_wrapper(captured_tool):
                async def on_invoke_tool_wrapper(tool_context: ToolContext[Any], tool_input: Any) -> Any:
                    await sequential_tools_lock.acquire()

                    async def invoke():
                        result = await captured_tool.on_invoke_tool(tool_context, tool_input)
                        # Ensure Pydantic objects are serialized to dict for LLM compatibility
                        if hasattr(result, "model_dump"):
                            return result.model_dump()
                        elif hasattr(result, "dict"):
                            return result.dict()
                        return result

                    try:
                        if disable_tool_autowrapping:
                            return await invoke()

                        ctx = current_context()
                        if ctx is None:
                            raise RuntimeError(
                                "No current Restate context found, make sure to run inside a Restate handler"
                            )
                        return await ctx.run_typed(captured_tool.name, invoke)
                    finally:
                        sequential_tools_lock.release()

                return on_invoke_tool_wrapper

            wrapped_tools.append(dataclasses.replace(tool, on_invoke_tool=create_wrapper(tool)))
        else:
            wrapped_tools.append(tool)

    handoffs_with_wrapped_tools = []
    for handoff in agent.handoffs:
        # recursively wrap tools in handoff agents
        handoffs_with_wrapped_tools.append(sequentialize_and_wrap_tools(handoff, disable_tool_autowrapping))  # type: ignore

    return agent.clone(
        tools=wrapped_tools,
        handoffs=handoffs_with_wrapped_tools,
    )



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

from contextlib import asynccontextmanager
from restate.extensions import contextvar

import restate

from restate import Context, Service, HarnessEnvironment, extensions
import pytest

# ----- Asyncio fixtures


@pytest.fixture(scope="session")
def anyio_backend():
    return "asyncio"


pytestmark = [
    pytest.mark.anyio,
]

# -------- Restate services and restate fixture

greeter = Service("greeter")


def magic_function():
    ctx = extensions.current_context()
    assert ctx is not None
    return ctx.request().id


@greeter.handler()
async def greet(ctx: Context, name: str) -> str:
    id = magic_function()
    return f"Hello {id}!"


# -- context manager


@contextvar
@asynccontextmanager
async def my_resource_manager():
    yield "hello"


@greeter.handler(invocation_context_managers=[my_resource_manager])
async def greet_with_cm(ctx: Context, name: str) -> str:
    return my_resource_manager.value


@pytest.fixture(scope="session")
async def restate_test_harness():
    async with restate.create_test_harness(
        restate.app([greeter]), restate_image="ghcr.io/restatedev/restate:latest"
    ) as harness:
        yield harness


# ----- Tests


async def test_greeter(restate_test_harness: HarnessEnvironment):
    greeting = await restate_test_harness.client.service_call(greet, arg="bob")
    assert greeting.startswith("Hello ")


async def test_greeter_with_cm(restate_test_harness: HarnessEnvironment):
    greeting = await restate_test_harness.client.service_call(greet_with_cm, arg="bob")
    assert greeting == "hello"



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
wrap the restate._internal.PyVM class
"""
# pylint: disable=E1101,R0917
# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods

from dataclasses import dataclass
import typing
from restate._internal import (
    PyVM,
    PyHeader,
    PyFailure,
    VMException,
    PySuspended,
    PyVoid,
    PyStateKeys,
    PyExponentialRetryConfig,
    PyDoProgressAnyCompleted,
    PyDoProgressReadFromInput,
    PyDoProgressExecuteRun,
    PyDoWaitForPendingRun,
    PyDoProgressCancelSignalReceived,
    CANCEL_NOTIFICATION_HANDLE,
)  # pylint: disable=import-error,no-name-in-module,line-too-long


@dataclass
class Invocation:
    """
    Invocation dataclass
    """

    invocation_id: str
    random_seed: int
    headers: typing.List[typing.Tuple[str, str]]
    input_buffer: bytes
    key: str


@dataclass
class RunRetryConfig:
    """
    Exponential Retry Configuration

    All duration/interval values are in milliseconds.
    """

    initial_interval: typing.Optional[int] = None
    max_attempts: typing.Optional[int] = None
    max_duration: typing.Optional[int] = None
    max_interval: typing.Optional[int] = None
    interval_factor: typing.Optional[float] = None


@dataclass
class Failure:
    """
    Failure
    """

    code: int
    message: str


@dataclass
class NotReady:
    """
    NotReady
    """


NOT_READY = NotReady()
CANCEL_HANDLE = CANCEL_NOTIFICATION_HANDLE

NotificationType = typing.Optional[typing.Union[bytes, Failure, NotReady, list[str], str]]


class Suspended:
    """
    Represents a suspended error
    """


SUSPENDED = Suspended()


class DoProgressAnyCompleted:
    """
    Represents a notification that any of the handles has completed.
    """


class DoProgressReadFromInput:
    """
    Represents a notification that the input needs to be read.
    """


class DoProgressExecuteRun:
    """
    Represents a notification that a run needs to be executed.
    """

    handle: int

    def __init__(self, handle):
        self.handle = handle


class DoProgressCancelSignalReceived:
    """
    Represents a notification that a cancel signal has been received
    """


class DoWaitPendingRun:
    """
    Represents a notification that a run is pending
    """


DO_PROGRESS_ANY_COMPLETED = DoProgressAnyCompleted()
DO_PROGRESS_READ_FROM_INPUT = DoProgressReadFromInput()
DO_PROGRESS_CANCEL_SIGNAL_RECEIVED = DoProgressCancelSignalReceived()
DO_WAIT_PENDING_RUN = DoWaitPendingRun()

DoProgressResult = typing.Union[
    DoProgressAnyCompleted,
    DoProgressReadFromInput,
    DoProgressExecuteRun,
    DoProgressCancelSignalReceived,
    DoWaitPendingRun,
]


# pylint: disable=too-many-public-methods
class VMWrapper:
    """
    A wrapper class for the restate_sdk._internal.PyVM class.
    It provides a type-friendly interface to our shared vm.
    """

    def __init__(self, headers: typing.List[typing.Tuple[str, str]]):
        self.vm = PyVM(headers)

    def get_response_head(self) -> typing.Tuple[int, typing.Iterable[typing.Tuple[str, str]]]:
        """
        Retrieves the response head from the virtual machine.

        Returns:
            A tuple containing the status code and a list of header tuples.
        """
        result = self.vm.get_response_head()
        return (result.status_code, result.headers)

    def notify_input(self, input_buf: bytes):
        """Send input to the virtual machine."""
        self.vm.notify_input(input_buf)

    def notify_input_closed(self):
        """Notify the virtual machine that the input has been closed."""
        self.vm.notify_input_closed()

    def notify_error(self, error: str, stacktrace: str):
        """Notify the virtual machine of an error."""
        self.vm.notify_error(error, stacktrace)

    def take_output(self) -> typing.Optional[bytes]:
        """Take the output from the virtual machine."""
        return self.vm.take_output()

    def is_ready_to_execute(self) -> bool:
        """Returns true when the VM is ready to operate."""
        return self.vm.is_ready_to_execute()

    def is_completed(self, handle: int) -> bool:
        """Returns true when the notification handle is completed and hasn't been taken yet."""
        return self.vm.is_completed(handle)

    # pylint: disable=R0911
    def do_progress(self, handles: list[int]) -> typing.Union[DoProgressResult, Exception, Suspended]:
        """Do progress with notifications."""
        try:
            result = self.vm.do_progress(handles)
        except VMException as e:
            return e
        if isinstance(result, PySuspended):
            return SUSPENDED
        if isinstance(result, PyDoProgressAnyCompleted):
            return DO_PROGRESS_ANY_COMPLETED
        if isinstance(result, PyDoProgressReadFromInput):
            return DO_PROGRESS_READ_FROM_INPUT
        if isinstance(result, PyDoProgressExecuteRun):
            return DoProgressExecuteRun(result.handle)
        if isinstance(result, PyDoProgressCancelSignalReceived):
            return DO_PROGRESS_CANCEL_SIGNAL_RECEIVED
        if isinstance(result, PyDoWaitForPendingRun):
            return DO_WAIT_PENDING_RUN
        return ValueError(f"Unknown progress type: {result}")

    def take_notification(self, handle: int) -> typing.Union[NotificationType, Exception, Suspended]:
        """Take the result of an asynchronous operation."""
        try:
            result = self.vm.take_notification(handle)
        except VMException as e:
            return e
        if isinstance(result, PySuspended):
            return SUSPENDED
        if result is None:
            return NOT_READY
        if isinstance(result, PyVoid):
            # success with an empty value
            return None
        if isinstance(result, bytes):
            # success with a non empty value
            return result
        if isinstance(result, PyStateKeys):
            # success with state keys
            return result.keys
        if isinstance(result, str):
            # success with invocation id
            return result
        if isinstance(result, PyFailure):
            # a terminal failure
            code = result.code
            message = result.message
            return Failure(code, message)
        return ValueError(f"Unknown result type: {result}")

    def sys_input(self) -> Invocation:
        """
        Retrieves the system input from the virtual machine.

        Returns:
            An instance of the Invocation class containing the system input.
        """
        inp = self.vm.sys_input()
        invocation_id: str = inp.invocation_id
        random_seed: int = inp.random_seed
        headers: typing.List[typing.Tuple[str, str]] = [(h.key, h.value) for h in inp.headers]
        input_buffer: bytes = bytes(inp.input)
        key: str = inp.key

        return Invocation(
            invocation_id=invocation_id, random_seed=random_seed, headers=headers, input_buffer=input_buffer, key=key
        )

    def sys_write_output_success(self, output: bytes):
        """
        Writes the output to the system.

        Args:
          output: The output to be written. It can be either a bytes or a Failure object.

        Returns:
            None
        """
        self.vm.sys_write_output_success(output)

    def sys_write_output_failure(self, output: Failure):
        """
        Writes the output to the system.

        Args:
          output: The output to be written. It can be either a bytes or a Failure object.

        Returns:
            None
        """
        res = PyFailure(output.code, output.message)
        self.vm.sys_write_output_failure(res)

    def sys_get_state(self, name) -> int:
        """
        Retrieves a key-value binding.

        Args:
            name: The name of the value to be retrieved.

        Returns:
            The value associated with the given name.
        """
        return self.vm.sys_get_state(name)

    def sys_get_state_keys(self) -> int:
        """
        Retrieves all keys.

        Returns:
            The state keys
        """
        return self.vm.sys_get_state_keys()

    def sys_set_state(self, name: str, value: bytes):
        """
        Sets a key-value binding.

        Args:
            name: The name of the value to be set.
            value: The value to be set.

        Returns:
            None
        """
        self.vm.sys_set_state(name, value)

    def sys_clear_state(self, name: str):
        """Clear the state associated with the given name."""
        self.vm.sys_clear_state(name)

    def sys_clear_all_state(self):
        """Clear the state associated with the given name."""
        self.vm.sys_clear_all_state()

    def sys_sleep(self, millis: int, name: typing.Optional[str] = None):
        """Ask to sleep for a given duration"""
        return self.vm.sys_sleep(millis, name)

    def sys_call(
        self,
        service: str,
        handler: str,
        parameter: bytes,
        key: typing.Optional[str] = None,
        idempotency_key: typing.Optional[str] = None,
        headers: typing.Optional[typing.List[typing.Tuple[str, str]]] = None,
    ):
        """Call a service"""
        if headers:
            headers = [PyHeader(key=h[0], value=h[1]) for h in headers]
        return self.vm.sys_call(service, handler, parameter, key, idempotency_key, headers)

    # pylint: disable=too-many-arguments
    def sys_send(
        self,
        service: str,
        handler: str,
        parameter: bytes,
        key: typing.Optional[str] = None,
        delay: typing.Optional[int] = None,
        idempotency_key: typing.Optional[str] = None,
        headers: typing.Optional[typing.List[typing.Tuple[str, str]]] = None,
    ) -> int:
        """
        send an invocation to a service, and return the handle
        to the promise that will resolve with the invocation id
        """
        if headers:
            headers = [PyHeader(key=h[0], value=h[1]) for h in headers]
        return self.vm.sys_send(service, handler, parameter, key, delay, idempotency_key, headers)

    def sys_run(self, name: str) -> int:
        """
        Register a run
        """
        return self.vm.sys_run(name)

    def sys_awakeable(self) -> typing.Tuple[str, int]:
        """
        Return a fresh awaitable
        """
        return self.vm.sys_awakeable()

    def sys_resolve_awakeable(self, name: str, value: bytes):
        """
        Resolve
        """
        self.vm.sys_complete_awakeable_success(name, value)

    def sys_reject_awakeable(self, name: str, failure: Failure):
        """
        Reject
        """
        py_failure = PyFailure(failure.code, failure.message)
        self.vm.sys_complete_awakeable_failure(name, py_failure)

    def propose_run_completion_success(self, handle: int, output: bytes) -> int:
        """
        Exit a side effect

        Args:
            output: The output of the side effect.

        Returns:
            handle
        """
        return self.vm.propose_run_completion_success(handle, output)

    def sys_get_promise(self, name: str) -> int:
        """Returns the promise handle"""
        return self.vm.sys_get_promise(name)

    def sys_peek_promise(self, name: str) -> int:
        """Peek into the workflow promise"""
        return self.vm.sys_peek_promise(name)

    def sys_complete_promise_success(self, name: str, value: bytes) -> int:
        """Complete the promise"""
        return self.vm.sys_complete_promise_success(name, value)

    def sys_complete_promise_failure(self, name: str, failure: Failure) -> int:
        """reject the promise on failure"""
        res = PyFailure(failure.code, failure.message)
        return self.vm.sys_complete_promise_failure(name, res)

    def propose_run_completion_failure(self, handle: int, output: Failure) -> int:
        """
        Exit a side effect

        Args:
            name: The name of the side effect.
            output: The output of the side effect.
        """
        res = PyFailure(output.code, output.message)
        return self.vm.propose_run_completion_failure(handle, res)

    # pylint: disable=line-too-long
    def propose_run_completion_transient(
        self, handle: int, failure: Failure, attempt_duration_ms: int, config: RunRetryConfig
    ):
        """
        Exit a side effect with a transient Error.
        This requires a retry policy to be provided.
        """
        py_failure = PyFailure(failure.code, failure.message)
        py_config = PyExponentialRetryConfig(
            config.initial_interval,
            config.max_attempts,
            config.max_duration,
            config.max_interval,
            config.interval_factor,
        )
        self.vm.propose_run_completion_failure_transient(handle, py_failure, attempt_duration_ms, py_config)

    def sys_end(self):
        """
        This method is responsible for ending the system.

        It calls the `sys_end` method of the `vm` object.
        """
        self.vm.sys_end()

    def sys_cancel(self, invocation_id: str):
        """
        Cancel a running invocation
        """
        self.vm.sys_cancel(invocation_id)

    def attach_invocation(self, invocation_id: str) -> int:
        """
        Attach to an invocation
        """
        return self.vm.attach_invocation(invocation_id)

    def is_replaying(self) -> bool:
        """Returns true if the state machine is replaying."""
        return self.vm.is_replaying()



# Copyright (c) 2025 - Restate Software, Inc., Restate GmbH
#
# This file is part of the Restate SDK for Python,
# which is released under the MIT license.
#
# You can find a copy of the license in file LICENSE in the root
# directory of this repository or package, or at
# https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
"""
Retry policy configuration for handler invocations exposed in the discovery manifest (protocol v4+).

Note: You can set these fields only if you register this service against restate-server >= 1.5
and discovery protocol v4. Otherwise, service discovery will fail.
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import Optional, Literal


@dataclass
class InvocationRetryPolicy:
    """
    Retry policy used by Restate when retrying failed handler invocations.

    Fields:
      - initial_interval: Initial delay before the first retry attempt.
      - exponentiation_factor: Exponential backoff multiplier used to compute the next retry delay.
      - max_interval: Upper bound for any computed retry delay.
      - max_attempts: Maximum number of attempts before giving up retrying.
            The initial call counts as the first attempt.
      - on_max_attempts: Behavior when reaching max attempts (pause or kill).
    """

    initial_interval: Optional[timedelta] = None
    exponentiation_factor: Optional[float] = None
    max_interval: Optional[timedelta] = None
    max_attempts: Optional[int] = None
    on_max_attempts: Optional[Literal["pause", "kill"]] = None



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

from .session import RestateSessionService
from .plugin import RestatePlugin

__all__ = [
    "RestateSessionService",
    "RestatePlugin",
]



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
contextvar utility for async context managers.
"""

import contextvars
from contextlib import asynccontextmanager
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    Callable,
    Generic,
    ParamSpec,
    TypeVar,
)

P = ParamSpec("P")
T = TypeVar("T")


class contextvar(Generic[P, T]):
    """
    A type-safe decorator for asynccontextmanager functions that captures the yielded value in a ContextVar.
    This is useful when integrating with frameworks that only support None yielded values from context managers.

    Example usage:
    ```python
    @contextvar
    @asynccontextmanager
    async def my_resource() -> AsyncIterator[str]:
        yield "hi"

    async def usage_example():
        async with my_resource():
            print(my_resource.value)  # prints "hi"
    ```


    """

    def __init__(self, func: Callable[P, AsyncContextManager[T]]):
        self.func = func
        self._value_var: contextvars.ContextVar[T | None] = contextvars.ContextVar("value")

    @property
    def value(self) -> T:
        """Return the value yielded by the wrapped context manager."""
        val = self._value_var.get()
        if val is None:
            raise LookupError(
                "Context manager value accessed outside of context manager scope (has not been entered yet)"
            )
        return val

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> AsyncContextManager[None]:
        @asynccontextmanager
        async def wrapper() -> AsyncGenerator[None, Any]:
            async with self.func(*args, **kwargs) as value:
                token = self._value_var.set(value)
                try:
                    yield  # we make it yield None, as the value is accessible via .value()
                finally:
                    self._value_var.reset(token)

        return wrapper()



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

# pylint: disable=R0917
# pylint: disable=R0801
"""
This represents common types used throughout the Restate SDK for Python.
"""

from dataclasses import dataclass
from types import UnionType
from typing import Any, Tuple, Literal, Union, get_args, get_origin

from restate.client_types import RestateClient


@dataclass
class HarnessEnvironment:
    """Information about the test environment"""

    ingress_url: str
    """The URL of the Restate ingress endpoint used in the test"""

    admin_api_url: str
    """The URL of the Restate admin API endpoint used in the test"""

    client: RestateClient
    """The Restate client connected to the ingress URL"""


def extract_core_type(annotation: Any) -> Tuple[Literal["optional", "simple"], Any]:
    """
    Extract the core type from a type annotation.

    Currently only supports Optional[T] types.
    """
    if annotation is None:
        return "simple", annotation

    origin = get_origin(annotation)
    args = get_args(annotation)

    if (origin is UnionType or Union) and len(args) == 2 and type(None) in args:
        non_none_type = args[0] if args[1] is type(None) else args[1]
        return "optional", non_none_type

    return "simple", annotation



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains the ASGI server for the restate framework."""

import asyncio
import logging
from typing import Dict, TypedDict, Literal

from restate.discovery import compute_discovery_json
from restate.endpoint import Endpoint
from restate.server_context import ServerInvocationContext, DisconnectedException
from restate.server_types import Receive, ReceiveChannel, RestateAppT, Scope, Send, binary_to_header, header_to_binary  # pylint: disable=line-too-long
from restate.vm import VMWrapper
from restate.aws_lambda import is_running_on_lambda, wrap_asgi_as_lambda_handler

from restate._internal import PyIdentityVerifier, IdentityVerificationException  # pylint: disable=import-error,no-name-in-module
from restate._internal import SDK_VERSION  # pylint: disable=import-error,no-name-in-module

logger = logging.getLogger(__name__)

X_RESTATE_SERVER = header_to_binary([("x-restate-server", f"restate-sdk-python/{SDK_VERSION}")])


async def send_status(send, receive, status_code: int):
    """respond with a status code"""
    await send({"type": "http.response.start", "status": status_code, "headers": X_RESTATE_SERVER})
    # For more info on why this loop, see ServerInvocationContext.leave()
    # pylint: disable=R0801
    while True:
        event = await receive()
        if event is None:
            break
        if event.get("type") == "http.disconnect":
            break
        if event.get("type") == "http.request" and event.get("more_body", False) is False:
            break
    await send({"type": "http.response.body"})


async def send404(send, receive):
    """respond with a 404"""
    await send_status(send, receive, 404)


async def send_discovery(scope: Scope, send: Send, endpoint: Endpoint):
    """respond with a discovery"""
    discovered_as: Literal["request_response", "bidi"]
    if scope["http_version"] == "1.1":
        discovered_as = "request_response"
    else:
        discovered_as = "bidi"

    # Extract Accept header from request
    accept_header = None
    for header_name, header_value in binary_to_header(scope["headers"]):
        if header_name.lower() == "accept":
            accept_header = header_value
            break

    # Negotiate discovery protocol version
    version = 2
    if accept_header:
        if "application/vnd.restate.endpointmanifest.v4+json" in accept_header:
            version = 4
        elif "application/vnd.restate.endpointmanifest.v3+json" in accept_header:
            version = 3
        elif "application/vnd.restate.endpointmanifest.v2+json" in accept_header:
            version = 2
        else:
            await send_status_with_error_text(send, 415, f"Unsupported discovery version ${accept_header}")
            return

    try:
        js = compute_discovery_json(endpoint, version, discovered_as)
        bin_headers = header_to_binary([("content-type", f"application/vnd.restate.endpointmanifest.v{version}+json")])
        bin_headers.extend(X_RESTATE_SERVER)
        await send({"type": "http.response.start", "status": 200, "headers": bin_headers, "trailers": False})
        await send(
            {
                "type": "http.response.body",
                "body": js.encode("utf-8"),
                "more_body": False,
            }
        )
    except ValueError as e:
        await send_status_with_error_text(send, 500, f"Error when computing discovery ${e}")
        return


async def send_status_with_error_text(send: Send, status_code: int, error_text: str):
    """respond with an health check"""
    headers = header_to_binary([("content-type", "text/plain")])
    headers.extend(X_RESTATE_SERVER)
    await send({"type": "http.response.start", "status": status_code, "headers": headers, "trailers": False})
    await send(
        {
            "type": "http.response.body",
            "body": error_text.encode("utf-8"),
            "more_body": False,
        }
    )


async def send_health_check(send: Send):
    """respond with an health check"""
    headers = header_to_binary([("content-type", "application/json")])
    headers.extend(X_RESTATE_SERVER)
    await send({"type": "http.response.start", "status": 200, "headers": headers, "trailers": False})
    await send(
        {
            "type": "http.response.body",
            "body": b'{"status":"ok"}',
            "more_body": False,
        }
    )


async def process_invocation_to_completion(
    vm: VMWrapper, handler, attempt_headers: Dict[str, str], receive: ReceiveChannel, send: Send
):
    """Invoke the user code."""
    status, res_headers = vm.get_response_head()
    res_bin_headers = header_to_binary(res_headers)
    res_bin_headers.extend(X_RESTATE_SERVER)
    await send({"type": "http.response.start", "status": status, "headers": res_bin_headers, "trailers": False})
    assert status == 200
    # ========================================
    # Read the input and the journal
    # ========================================
    while True:
        message = await receive()
        if message.get("type") == "http.disconnect":
            # everything ends here really ...
            return
        if message.get("type") == "http.request":
            body = message.get("body", None)
            assert isinstance(body, bytes)
            vm.notify_input(body)
        if not message.get("more_body", False):
            vm.notify_input_closed()
            break
        if vm.is_ready_to_execute():
            break
    # ========================================
    # Execute the user code
    # ========================================
    invocation = vm.sys_input()
    context = ServerInvocationContext(
        vm=vm, handler=handler, invocation=invocation, attempt_headers=attempt_headers, send=send, receive=receive
    )
    try:
        await context.enter()
    except asyncio.exceptions.CancelledError:
        context.on_attempt_finished()
        raise
    except DisconnectedException:
        # The client disconnected before we could send the response
        context.on_attempt_finished()
        return
    # pylint: disable=W0718
    except Exception:
        logger.exception("Exception in Restate handler")
    try:
        await context.leave()
    finally:
        context.on_attempt_finished()


class LifeSpanNotImplemented(ValueError):
    """Signal to the asgi server that we didn't implement lifespans"""


class ParsedPath(TypedDict):
    """Parsed path from the request."""

    type: Literal["invocation", "health", "discover", "unknown"]
    service: str | None
    handler: str | None


def parse_path(request: str) -> ParsedPath:
    """Parse the path from the request."""
    # The following routes are possible
    # $mountpoint/health
    # $mountpoint/discover
    # $mountpoint/invoke/:service/:handler
    # as we don't know the mountpoint, we need to check the path carefully
    fragments = request.rsplit("/", 4)
    # /invoke/:service/:handler
    if len(fragments) >= 3 and fragments[-3] == "invoke":
        return {"type": "invocation", "handler": fragments[-1], "service": fragments[-2]}
    # /health
    if fragments[-1] == "health":
        return {"type": "health", "service": None, "handler": None}
    # /discover
    if fragments[-1] == "discover":
        return {"type": "discover", "service": None, "handler": None}
    # anything other than invoke is 404
    return {"type": "unknown", "service": None, "handler": None}


def asgi_app(endpoint: Endpoint) -> RestateAppT:
    """Create an ASGI-3 app for the given endpoint."""

    # Prepare request signer
    identity_verifier = PyIdentityVerifier(endpoint.identity_keys)

    async def app(scope: Scope, receive: Receive, send: Send):
        try:
            if scope["type"] == "lifespan":
                raise LifeSpanNotImplemented()
            if scope["type"] != "http":
                raise NotImplementedError(f"Unknown scope type {scope['type']}")

            request_path = scope["path"]
            assert isinstance(request_path, str)
            request: ParsedPath = parse_path(request_path)

            # Health check
            if request["type"] == "health":
                await send_health_check(send)
                return

            # Verify Identity
            assert not isinstance(scope["headers"], str)
            assert hasattr(scope["headers"], "__iter__")
            request_headers = binary_to_header(scope["headers"])
            try:
                identity_verifier.verify(request_headers, request_path)
            except IdentityVerificationException:
                # Identify verification failed, send back unauthorized and close
                await send_status(send, receive, 401)
                return

            # might be a discovery request
            if request["type"] == "discover":
                await send_discovery(scope, send, endpoint)
                return
            # anything other than invoke is 404
            if request["type"] == "unknown":
                await send404(send, receive)
                return
            assert request["type"] == "invocation"
            assert request["service"] is not None
            assert request["handler"] is not None
            service_name, handler_name = request["service"], request["handler"]
            service = endpoint.services.get(service_name)
            if not service:
                await send404(send, receive)
                return
            handler = service.handlers.get(handler_name)
            if not handler:
                await send404(send, receive)
                return
            #
            # At this point we have a valid handler.
            # Let us setup restate's execution context for this invocation and handler.
            #
            receive_channel = ReceiveChannel(receive)
            try:
                await process_invocation_to_completion(
                    VMWrapper(request_headers), handler, dict(request_headers), receive_channel, send
                )
            finally:
                await receive_channel.close()
        except LifeSpanNotImplemented as e:
            raise e
        except Exception as e:
            logger.exception("Exception in ASGI app")
            raise e

    if is_running_on_lambda():
        # If we're on Lambda, just return the adapter
        return wrap_asgi_as_lambda_handler(app)

    return app



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=C0301

from datetime import timedelta

from restate import Workflow, WorkflowContext, WorkflowSharedContext
from restate import select
from restate import TerminalError

TIMEOUT = timedelta(seconds=10)

payment = Workflow("payment")


@payment.main()
async def pay(ctx: WorkflowContext, amount: int):
    workflow_key = ctx.key()
    ctx.set("status", "verifying payment")

    # Call the payment service
    def payment_gateway():
        print("Please approve this payment: ")
        print("To approve use:")
        print(f"""curl http://localhost:8080/payment/{workflow_key}/payment_verified --json '"approved"' """)
        print("")
        print("To decline use:")
        print(f"""curl http://localhost:8080/payment/{workflow_key}/payment_verified --json '"declined"' """)

    await ctx.run_typed("payment", payment_gateway)

    ctx.set("status", "waiting for the payment provider to approve")

    # Wait for the payment to be verified

    match await select(result=ctx.promise("verify.payment").value(), timeout=ctx.sleep(TIMEOUT)):
        case ["result", "approved"]:
            ctx.set("status", "payment approved")
            return {"success": True}
        case ["result", "declined"]:
            ctx.set("status", "payment declined")
            raise TerminalError(message="Payment declined", status_code=401)
        case ["timeout", _]:
            ctx.set("status", "payment verification timed out")
            raise TerminalError(message="Payment verification timed out", status_code=410)


@payment.handler()
async def payment_verified(ctx: WorkflowSharedContext, result: str):
    promise = ctx.promise("verify.payment", type_hint=str)
    await promise.resolve(result)



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

from contextlib import asynccontextmanager
import restate
from restate import (
    Context,
    RunOptions,
    Service,
    TerminalError,
    VirtualObject,
    Workflow,
    WorkflowContext,
)
from restate.serde import DefaultSerde
import pytest
import typing

# ----- Asyncio fixtures


@pytest.fixture(scope="session")
def anyio_backend():
    return "asyncio"


pytestmark = [
    pytest.mark.anyio,
]


def ohoh():
    raise TerminalError("Simulated terminal error")


@asynccontextmanager
async def simple_harness(service: Service | VirtualObject | Workflow) -> typing.AsyncIterator[restate.RestateClient]:
    async with restate.create_test_harness(
        restate.app([service]), restate_image="ghcr.io/restatedev/restate:latest"
    ) as restate_test_harness:
        yield restate_test_harness.client


async def test_sanity():
    greeter = Service("greeter")

    @greeter.handler()
    async def greet(ctx: Context, name: str) -> str:
        await ctx.run_typed("foo", ohoh, RunOptions(max_attempts=3))
        return "hi"

    async with simple_harness(greeter) as client:
        with pytest.raises(Exception):
            await client.service_call(greet, arg="bob")


async def test_wrapped_terminal_exception():
    greeter = Service("greeter")

    @greeter.handler()
    async def greet(ctx: Context, name: str) -> str:
        try:
            await ctx.run_typed("foo", ohoh)
            return "hi"
        except TerminalError as te:
            raise ValueError("Wrapped terminal error") from te

    async with simple_harness(greeter) as client:
        with pytest.raises(Exception):
            await client.service_call(greet, arg="bob")


async def test_promise_default_serde():
    workflow = Workflow("test_workflow")

    @workflow.main()
    async def run(ctx: WorkflowContext) -> str:
        promise = ctx.promise("test.promise", type_hint=str)

        assert isinstance(promise.serde, DefaultSerde), f"Expected DefaultSerde but got {type(promise.serde).__name__}"

        await promise.resolve("success")
        return await promise.value()

    async with simple_harness(workflow) as client:
        result = await client.workflow_call(run, key="test-key", arg=None)
        assert result == "success"


async def test_handler_with_union_none():
    greeter = Service("greeter")

    @greeter.handler()
    async def greet(ctx: Context, name: str) -> str | None:
        return "hi"

    async with simple_harness(greeter) as client:
        res = await client.service_call(greet, arg="bob")
        assert res == "hi"


async def test_handler_with_ctx_none():
    greeter = Service("greeter")

    async def maybe_something() -> str | None:
        return "hi"

    @greeter.handler()
    async def greet(ctx: Context, name: str) -> str | None:
        return await ctx.run_typed("foo", maybe_something)

    async with simple_harness(greeter) as client:
        res = await client.service_call(greet, arg="bob")
        assert res == "hi"



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
ADK session service implementation using Restate Virtual Objects as the backing store.
"""

import restate

from typing import Optional, Any, cast

from google.adk.sessions import Session
from google.adk.events.event import Event
from google.adk.sessions.base_session_service import (
    BaseSessionService,
    ListSessionsResponse,
    GetSessionConfig,
)

from restate.extensions import current_context


class RestateSessionService(BaseSessionService):
    def ctx(self) -> restate.ObjectContext:
        return cast(restate.ObjectContext, current_context())

    async def create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        if session_id is None:
            session_id = str(self.ctx().uuid())

        session = await self.ctx().get(f"session_store::{session_id}", type_hint=Session) or Session(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
            state=state or {},
        )
        self.ctx().set(f"session_store::{session_id}", session)
        return session

    async def get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        # TODO : Handle config options
        return await self.ctx().get(f"session_store::{session_id}", type_hint=Session) or Session(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
        )

    async def list_sessions(self, *, app_name: str, user_id: Optional[str] = None) -> ListSessionsResponse:
        state_keys = await self.ctx().state_keys()
        sessions = []
        for key in state_keys:
            if key.startswith("session_store::"):
                session = await self.ctx().get(key, type_hint=Session)
                if session is not None:
                    sessions.append(session)
        return ListSessionsResponse(sessions=sessions)

    async def delete_session(self, *, app_name: str, user_id: str, session_id: str) -> None:
        self.ctx().clear(f"session_store::{session_id}")

    async def append_event(self, session: Session, event: Event) -> Event:
        """Appends an event to a session object."""
        if event.partial:
            return event
        # For now, we also store temp state
        event = self._trim_temp_delta_state(event)
        self._update_session_state(session, event)
        session.events.append(event)
        return event


async def flush_session_state(ctx: restate.ObjectContext, session: Session):
    session_to_store = session.model_copy()
    # Remove restate-specific context that got added by the plugin before storing
    session_to_store.state.pop("restate_context", None)
    deterministic_session = await ctx.run_typed(
        "store session", lambda: session_to_store, restate.RunOptions(type_hint=Session)
    )
    ctx.set(f"session_store::{session.id}", deterministic_session)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the Lambda/ASGI adapter.
"""

import asyncio
import base64
import os
from typing import cast, Union, Any

from restate.server_types import (
    ASGIApp,
    Receive,
    RestateLambdaHandler,
    Scope,
    HTTPResponseStartEvent,
    HTTPResponseBodyEvent,
    HTTPRequestEvent,
    RestateLambdaRequest,
    RestateLambdaResponse,
)


def create_scope(req: RestateLambdaRequest) -> Scope:
    """
    Create ASGI scope from lambda request
    """
    headers = {k.lower(): v for k, v in req.get("headers", {}).items()}
    http_method = req["httpMethod"]
    path = req["path"]

    return {
        "type": "http",
        "method": http_method,
        "http_version": "1.1",
        "headers": [(k.encode(), v.encode()) for k, v in headers.items()],
        "path": path,
        "scheme": headers.get("x-forwarded-proto", "https"),
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "raw_path": path.encode(),
        "root_path": "",
        "query_string": b"",
        "client": None,
        "server": None,
        "extensions": None,
    }


def request_to_receive(req: RestateLambdaRequest) -> Receive:
    """
    Create ASGI Receive from lambda request
    """
    assert req["isBase64Encoded"]
    body = base64.b64decode(req["body"])

    events = cast(
        list[HTTPRequestEvent],
        [
            {"type": "http.request", "body": body, "more_body": False},
            {"type": "http.request", "body": b"", "more_body": False},
        ],
    )

    async def recv() -> HTTPRequestEvent:
        if len(events) != 0:
            return events.pop(0)
        # If we are out of events, return a future that will never complete
        f = asyncio.Future[HTTPRequestEvent]()
        return await f

    return recv


class ResponseCollector:
    """
    Response collector from ASGI Send to Lambda
    """

    def __init__(self):
        self.body = bytearray()
        self.headers = {}
        self.status_code = 500

    async def __call__(self, message: Union[HTTPResponseStartEvent, HTTPResponseBodyEvent]) -> None:
        """
        Implements ASGI send contract
        """
        if message["type"] == "http.response.start":
            self.status_code = cast(int, message["status"])
            self.headers = {key.decode("utf-8"): value.decode("utf-8") for key, value in message["headers"]}
        elif message["type"] == "http.response.body" and "body" in message:
            self.body.extend(message["body"])
        return

    def to_lambda_response(self) -> RestateLambdaResponse:
        """
        Convert collected values to lambda response
        """
        return {
            "statusCode": self.status_code,
            "headers": self.headers,
            "isBase64Encoded": True,
            "body": base64.b64encode(self.body).decode(),
        }


def is_running_on_lambda() -> bool:
    """
    :return: true if this Python script is running on Lambda
    """
    # https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html
    return "AWS_LAMBDA_FUNCTION_NAME" in os.environ


def wrap_asgi_as_lambda_handler(asgi_app: ASGIApp) -> RestateLambdaHandler:
    """
    Wrap the given asgi_app in a Lambda handler
    """
    # Setup AsyncIO
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    def lambda_handler(event: RestateLambdaRequest, _context: Any) -> RestateLambdaResponse:
        loop = asyncio.get_event_loop()

        scope = create_scope(event)
        recv = request_to_receive(event)
        send = ResponseCollector()

        asgi_instance = asgi_app(scope, recv, send)
        asgi_task = loop.create_task(asgi_instance)  # type: ignore[var-annotated, arg-type]
        loop.run_until_complete(asgi_task)

        return send.to_lambda_response()

    return lambda_handler



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""Verification test"""

from datetime import timedelta
import json
from typing import TypedDict
import typing
import random


from restate.context import Context, ObjectContext, ObjectSharedContext
from restate.exceptions import TerminalError
from restate.object import VirtualObject
from restate.serde import JsonSerde

import restate

SET_STATE = 1
GET_STATE = 2
CLEAR_STATE = 3
INCREMENT_STATE_COUNTER = 4
INCREMENT_STATE_COUNTER_INDIRECTLY = 5
SLEEP = 6
CALL_SERVICE = 7
CALL_SLOW_SERVICE = 8
INCREMENT_VIA_DELAYED_CALL = 9
SIDE_EFFECT = 10
THROWING_SIDE_EFFECT = 11
SLOW_SIDE_EFFECT = 12
RECOVER_TERMINAL_CALL = 13
RECOVER_TERMINAL_MAYBE_UN_AWAITED = 14
AWAIT_PROMISE = 15
RESOLVE_AWAKEABLE = 16
REJECT_AWAKEABLE = 17
INCREMENT_STATE_COUNTER_VIA_AWAKEABLE = 18
CALL_NEXT_LAYER_OBJECT = 19

# suppress missing docstring
# pylint: disable=C0115
# pylint: disable=C0116
# pylint: disable=C0301
# pylint: disable=R0914, R0912, R0915, R0913


helper = restate.Service("ServiceInterpreterHelper")


@helper.handler()
async def ping(ctx: Context) -> None:  # pylint: disable=unused-argument
    pass


@helper.handler()
async def echo(ctx: Context, parameters: str) -> str:  # pylint: disable=unused-argument
    return parameters


@helper.handler(name="echoLater")
async def echo_later(ctx: Context, parameter: dict[str, typing.Any]) -> str:
    await ctx.sleep(timedelta(milliseconds=parameter["sleep"]))
    return parameter["parameter"]


@helper.handler(name="terminalFailure")
async def terminal_failure(ctx: Context) -> str:
    raise TerminalError("bye")


@helper.handler(name="incrementIndirectly")
async def increment_indirectly(ctx: Context, parameter) -> None:
    layer = parameter["layer"]
    key = parameter["key"]

    program = {
        "commands": [
            {
                "kind": INCREMENT_STATE_COUNTER,
            },
        ],
    }

    program_bytes = json.dumps(program).encode("utf-8")

    ctx.generic_send(f"ObjectInterpreterL{layer}", "interpret", program_bytes, key)


@helper.handler(name="resolveAwakeable")
async def resolve_awakeable(ctx: Context, aid: str) -> None:
    ctx.resolve_awakeable(aid, "ok")


@helper.handler(name="rejectAwakeable")
async def reject_awakeable(ctx: Context, aid: str) -> None:
    ctx.reject_awakeable(aid, "error")


@helper.handler(name="incrementViaAwakeableDance")
async def increment_via_awakeable_dance(ctx: Context, input: dict[str, typing.Any]) -> None:
    tx_promise_id = input["txPromiseId"]
    layer = input["interpreter"]["layer"]
    key = input["interpreter"]["key"]

    aid, promise = ctx.awakeable()
    ctx.resolve_awakeable(tx_promise_id, aid)
    await promise

    program = {
        "commands": [
            {
                "kind": INCREMENT_STATE_COUNTER,
            },
        ],
    }

    program_bytes = json.dumps(program).encode("utf-8")

    ctx.generic_send(f"ObjectInterpreterL{layer}", "interpret", program_bytes, key)


class SupportService:
    def __init__(self, ctx: ObjectContext) -> None:
        self.ctx = ctx
        self.serde = JsonSerde[typing.Any]()

    async def call(self, method: str, arg: typing.Any) -> typing.Any:
        buffer = self.serde.serialize(arg)
        out_buffer = await self.ctx.generic_call("ServiceInterpreterHelper", method, buffer)
        return self.serde.deserialize(out_buffer)

    def send(self, method: str, arg: typing.Any, delay: int | None = None) -> None:
        buffer = self.serde.serialize(arg)
        if delay is None:
            send_delay = None
        else:
            send_delay = timedelta(milliseconds=delay)
        self.ctx.generic_send("ServiceInterpreterHelper", method, buffer, send_delay=send_delay)

    async def ping(self) -> None:
        return await self.call(method="ping", arg=None)

    async def echo(self, parameters: str) -> str:
        return await self.call(method="echo", arg=parameters)

    async def echo_later(self, parameter: str, sleep: int) -> str:
        arg = {"parameter": parameter, "sleep": sleep}
        return await self.call(method="echoLater", arg=arg)

    async def terminal_failure(self) -> str:
        return await self.call(method="terminalFailure", arg=None)

    async def increment_indirectly(self, layer: int, key: str, delay: typing.Optional[int] = None) -> None:
        arg = {"layer": layer, "key": key}
        self.send(method="incrementIndirectly", arg=arg, delay=delay)

    def resolve_awakeable(self, aid: str) -> None:
        self.send("resolveAwakeable", aid)

    def reject_awakeable(self, aid: str) -> None:
        self.send("rejectAwakeable", aid)

    def increment_via_awakeable_dance(self, layer: int, key: str, tx_promise_id: str) -> None:
        arg = {"interpreter": {"layer": layer, "key": key}, "txPromiseId": tx_promise_id}
        self.send("incrementViaAwakeableDance", arg)


class Command(TypedDict):
    kind: int
    key: int
    duration: int
    sleep: int
    index: int
    program: typing.Any  # avoid circular type


Program = dict[typing.Literal["commands"], typing.List[Command]]


async def interpreter(layer: int, ctx: ObjectContext, program: Program) -> None:
    """Interprets a command and executes it."""
    service = SupportService(ctx)
    coros: dict[int, typing.Tuple[typing.Any, typing.Awaitable[typing.Any]]] = {}

    async def await_promise(index: int) -> None:
        if index not in coros:
            return

        expected, coro = coros[index]
        del coros[index]
        try:
            result = await coro
        except TerminalError:
            result = "rejected"

        if result != expected:
            raise TerminalError(f"Expected {expected} but got {result}")

    for i, command in enumerate(program["commands"]):
        command_type = command["kind"]
        if command_type == SET_STATE:
            ctx.set(f"key-{command['key']}", f"value-{command['key']}")
        elif command_type == GET_STATE:
            await ctx.get(f"key-{command['key']}")
        elif command_type == CLEAR_STATE:
            ctx.clear(f"key-{command['key']}")
        elif command_type == INCREMENT_STATE_COUNTER:
            c = await ctx.get("counter") or 0
            c += 1
            ctx.set("counter", c)
        elif command_type == SLEEP:
            duration = timedelta(milliseconds=command["duration"])
            await ctx.sleep(duration)
        elif command_type == CALL_SERVICE:
            expected = f"hello-{i}"
            coros[i] = (expected, service.echo(expected))
        elif command_type == INCREMENT_VIA_DELAYED_CALL:
            delay = command["duration"]
            await service.increment_indirectly(layer=layer, key=ctx.key(), delay=delay)
        elif command_type == CALL_SLOW_SERVICE:
            expected = f"hello-{i}"
            coros[i] = (expected, service.echo_later(expected, command["sleep"]))
        elif command_type == SIDE_EFFECT:
            expected = f"hello-{i}"
            result = await ctx.run_typed("sideEffect", lambda: expected)
            if result != expected:
                raise TerminalError(f"Expected {expected} but got {result}")
        elif command_type == SLOW_SIDE_EFFECT:
            pass
        elif command_type == RECOVER_TERMINAL_CALL:
            try:
                await service.terminal_failure()
            except TerminalError:
                pass
            else:
                raise TerminalError("Expected terminal error")
        elif command_type == RECOVER_TERMINAL_MAYBE_UN_AWAITED:
            pass
        elif command_type == THROWING_SIDE_EFFECT:

            async def side_effect():
                if bool(random.getrandbits(1)):
                    raise ValueError("Random error")

            await ctx.run_typed("throwingSideEffect", side_effect)
        elif command_type == INCREMENT_STATE_COUNTER_INDIRECTLY:
            await service.increment_indirectly(layer=layer, key=ctx.key())
        elif command_type == AWAIT_PROMISE:
            index = command["index"]
            await await_promise(index)
        elif command_type == RESOLVE_AWAKEABLE:
            name, promise = ctx.awakeable()
            coros[i] = ("ok", promise)
            service.resolve_awakeable(name)
        elif command_type == REJECT_AWAKEABLE:
            name, promise = ctx.awakeable()
            coros[i] = ("rejected", promise)
            service.reject_awakeable(name)
        elif command_type == INCREMENT_STATE_COUNTER_VIA_AWAKEABLE:
            tx_promise_id, tx_promise = ctx.awakeable()
            service.increment_via_awakeable_dance(layer=layer, key=ctx.key(), tx_promise_id=tx_promise_id)
            their_promise_for_us_to_resolve: str = await tx_promise
            ctx.resolve_awakeable(their_promise_for_us_to_resolve, "ok")
        elif command_type == CALL_NEXT_LAYER_OBJECT:
            next_layer = f"ObjectInterpreterL{layer + 1}"
            key = f"{command['key']}"
            program = command["program"]
            js_program = json.dumps(program)
            raw_js_program = js_program.encode("utf-8")
            promise = ctx.generic_call(next_layer, "interpret", raw_js_program, key)
            coros[i] = (b"", promise)
        else:
            raise ValueError(f"Unknown command type: {command_type}")
        await await_promise(i)


def make_layer(i):
    layer = VirtualObject(f"ObjectInterpreterL{i}")

    @layer.handler()
    async def interpret(ctx: ObjectContext, program: Program) -> None:
        await interpreter(i, ctx, program)

    @layer.handler(kind="shared")
    async def counter(ctx: ObjectSharedContext) -> int:
        return await ctx.get("counter") or 0

    return layer


layer_0 = make_layer(0)
layer_1 = make_layer(1)
layer_2 = make_layer(2)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains internal extensions apis"""

from .server_context import current_context
from .context_managers import contextvar

__all__ = ["current_context", "contextvar"]
